"""
Vtrace Debugger Framework

Vtrace is a *mostly* native python debugging framework which
can be used to quickly write programatic debuggers and research
tools.

I'm not known for writting great docs...  but the code should
be pretty straight forward...

This has been in use for many years privately, but is nowhere
*near* free of bugs...  idiosyncracies abound.

==== Werd =====================================================

Blah blah blah... many more docs to come.

Brought to you by kenshoto.  e-mail invisigoth.

Greetz:
    h1kari - eeeeeooorrrmmm  CHKCHKCHKCHKCHKCHKCHK
    Ghetto - wizoo... to the tizoot.
    atlas - *whew* finally...  no more teasing...
    beatle/dnm - come out and play yo!
    The Kenshoto Gophers.
    Blackhats Everywhere.

"""
# Copyright (C) 2007 Invisigoth - See LICENSE file for details
import os
import re
import code
import time
import types
import logging
import platform

import envi
import envi.bits as e_bits
import envi.memory as e_mem
import envi.registers as e_reg
import envi.expression as e_expr
import envi.symstore.resolver as e_resolv
import envi.symstore.symcache as e_symcache

import vstruct
from vtrace.const import *

logger = logging.getLogger(__name__)
remote = None       # If set, we're a vtrace client (set to serverhost)
cobra_daemon = None
port = 0x5656
verbose = False

# File Descriptor / Handle Types
FD_UNKNOWN = 0 # Unknown or we don't have a type for it
FD_FILE = 1
FD_SOCKET = 2
FD_PIPE = 3
FD_LOCK = 4   # Win32 Mutant/Lock/Semaphore
FD_EVENT = 5  # Win32 Event/KeyedEvent
FD_THREAD = 6 # Win32 Thread
FD_REGKEY = 7 # Win32 Registry Key

# Vtrace Symbol Types
SYM_MISC = -1
SYM_GLOBAL = 0 # Global (mostly vars)
SYM_LOCAL = 1 # Locals
SYM_FUNCTION = 2 # Functions
SYM_SECTION = 3 # Binary section
SYM_META = 4 # Info that we enumerate

# Vtrace Symbol Offsets
VSYM_NAME = 0
VSYM_ADDR = 1
VSYM_SIZE = 2
VSYM_TYPE = 3
VSYM_FILE = 4

from vtrace.rmi import *
from vtrace.notifiers import *
from vtrace.breakpoints import *
from vtrace.watchpoints import *
import vtrace.util as v_util

class PlatformException(Exception):
    """
    A universal way to represent a failure in the
    platform layer for this tracer.  platformFoo methods
    should raise this rather than allowing their platform
    specific exception types (which don't likely pickle, or
    are not cross platform)
    """
    pass

class AccessViolation(Exception):
    """
    An exception which is raised on bad-touch to memory
    """
    def __init__(self, va, perm=0):
        self.va = va
        self.perm = perm
        Exception.__init__(self, "AccessViolation at 0x%.8x (%d)" % (va, perm))

class Trace(e_mem.IMemory, e_reg.RegisterContext, e_resolv.SymbolResolver, object):
    """
    The main tracer object.  A trace instance is dynamically generated using
    this and *many* potential mixin classes.  However, API users should *not*
    worry about the methods that come from the mixins...  Everything that is
    *meant* to be used from the API is contained and documented here.
    """
    def __init__(self, archname=None):
        # For the crazy thread-call-proxy-thing
        # (must come first for __getattribute__
        self.requires_thread = {}
        self.proxymeth = None  # FIXME hack for now...
        self._released = False

        # The universal place for all modes
        # that might be platform dependant...
        self.modes = {}
        self.modedocs = {}
        self.notifiers = {}
        self.namecache = []

        # For all transient data (if notifiers want
        # to track stuff per-trace
        self.metadata = {}

        self.initMode("RunForever", False, "Run until RunForever = False")
        self.initMode("NonBlocking", False, "A call to wait() fires a thread to wait *for* you")
        self.initMode("ThreadProxy", True, "Proxy necessary requests through a single thread (can deadlock...)")
        self.initMode("SingleStep", False, "All calls to run() actually just step.  This allows RunForever + SingleStep to step forever ;)")
        self.initMode("FastStep", False, "All stepi() will NOT generate a step event")

        self.regcache = None
        self.regcachedirty = False
        self.sus_threads = {}   # A dictionary of suspended threads

        # Set if we're a server and this trace is proxied
        self.proxy = None

        # Set us up with an envi arch module
        # FIXME eventually we should just inherit one...
        if archname is None:
            archname = envi.getCurrentArch()

        arch = envi.getArchByName(archname)
        self.setMeta('Architecture', archname)
        self.arch = envi.getArchModule(name=archname)

        e_resolv.SymbolResolver.__init__(self, width=self.arch.getPointerSize())
        e_mem.IMemory.__init__(self, arch=arch)
        e_reg.RegisterContext.__init__(self)

        # Add event numbers to here for auto-continue
        self.auto_continue = [NOTIFY_LOAD_LIBRARY, NOTIFY_CREATE_THREAD, NOTIFY_UNLOAD_LIBRARY, NOTIFY_EXIT_THREAD, NOTIFY_DEBUG_PRINT]

        # Create a LoadLibrary hook to enable simple and consistent 
        # Break-On-Load/Init functionality. This is also necessary for
        # resolving symbols on new library loads.
        self.registerNotifier(NOTIFY_LOAD_LIBRARY, LibraryNotifier())

    def setBreakOnLibraryLoad(self, setting=True):
        '''
        Cause execution to halt when a new library is loaded.
        '''
        logger.info("setting 'BreakOnLibraryLoad' to %r", setting)
        self.setMeta('BreakOnLibraryLoad', setting) 

    def setBreakOnLibraryInit(self, setting=True):
        '''
        Set breakpoint on a newly loaded Library's init function 
        (aka. <Libname>.__entry)
        '''
        logger.info("setting 'BreakOnLibraryInit' to %r", setting)
        self.setMeta('BreakOnLibraryInit', setting)

    def execute(self, cmdline):
        """
        Start a new process and debug it
        """
        if self.isAttached():
            raise Exception("ERROR - Tracer must first be detached before you can execute()")

        pid = self.platformExec(cmdline)
        self._justAttached(pid)
        self.setMeta('ExecCommand', cmdline)
        self.wait()

    def parseOpcodes(self, num, va=None):
        '''
        Returns next num of linear disasm'd opcodes objects.  Optionally pass
        a va to start there instead of the current program counter.
        '''
        if num <= 0:
            raise Exception('you must specify a positive number of opcodes')

        if va is None:
            va = self.getProgramCounter()

        ops = []
        for i in range(0, num):
            op = self.parseOpcode(va)
            ops.append(op)
            va += op.size

        return ops

    def getCurrentSignal(self):
        '''
        Retrieve the current signal/exception posted to the process.
        If there are no pending signals/exceptions the API will return
        None.  For POSIX systems, this will be a traditional POSIX signal.
        For Windows systems it will be a current exception code (if any).

        Example: sig = trace.getCurrentSignal()
        '''
        return self.platformGetSignal()

    def setCurrentSignal(self, sig=None):
        '''
        Set the currently pending signal for delivery to the target process on
        continue.  This is intended for use by programs wishing the mask or
        change the delivery of exceptions on a NOTIFY_SIGNAL event.

        Example:  trace.setCurrentSignal(None)
        '''
        return self.platformSetSignal(sig)

    def addIgnoreSignal(self, code, address=0):
        """
        By adding an IgnoreSignal you tell the tracer object to
        supress the notification of a particular type of signal.
        In POSIX, these are regular signals, in Win32, these
        are exception codes.  This is mostly useful in RunForever
        mode because you still need the process to begin running again.
        (these may be viewed/modified by the metadata key "IgnoredSignals")
        FIXME: make address do something.
        """
        self.getMeta("IgnoredSignals").append(code)

    def delIgnoreSignal(self, code, address=0):
        """
        See addIgnoreSignal for a description of signal ignoring.
        This removes an ignored signal and re-enables it's delivery.
        """
        self.getMeta("IgnoredSignals").remove(code)

    def attach(self, pid):
        """
        Attach to a new process ID.
        """
        if self.isAttached():
            self.detach()

        try:
            self.platformAttach(pid)
            self._justAttached(pid)
            self.wait()
        except Exception as msg:
            raise PlatformException(str(msg))

    def stepi(self):
        """
        Single step the target process ONE instruction (and do
        NOT activate breakpoints for the one step). Also, we
        don't deliver pending signals for the single step...
        Use the mode FastStep to allow/suppress notifier callbacks on step
        """
        self.requireNotRunning()

        # Since we don't go through the normal run/wait
        # code, we have a little house-keeping to do...
        self.curbp = None

        self._syncRegs()
        self.platformStepi()
        event = self.platformWait()
        self.platformProcessEvent(event)

    def run(self, until=None):
        """
        Allow the traced target to continue execution.  (Depending on the mode
        "Blocking" this will either block until an event, or return immediately)
        Additionally, the argument until may be used to cause execution to
        continue until the specified address is reached (internally uses and
        removes a breakpoint).
        """
        self.requireAttached()
        self.requireNotRunning()
        self.requireNotExited()

        if self.getMode("SingleStep", False):
            self.steploop()

        else:
            if until is not None:
                self.setMode("RunForever", True)
                self.addBreakpoint(StopAndRemoveBreak(until))

            self._doRun()
            self.wait()

    def runAgain(self, val=True):
        """
        The runAgain() method may be used from inside a notifier
        (Notifier, Breakpoint, Watchpoint, etc...) to inform the trace
        that once event processing is complete, it should continue
        running the trace.
        """
        self.runagain = val

    def kill(self):
        """
        Kill the target process for this trace (will result in process
        exit and fire appropriate notifiers)
        """
        self.requireAttached()
        self.requireNotExited()
        self.platformKill()
        self.attached = False

    def detach(self):
        '''
        Detach from the currently attached process.
        '''
        self.requireNotRunning()
        self.fireNotifiers(NOTIFY_DETACH)
        self._syncRegs()
        self.platformDetach()
        self.attached = False
        self.pid = 0
        self.mapcache = None

    def release(self):
        '''
        Release resources for this tracer.  This API should be called
        once you are done with the trace.
        '''
        if not self._released:
            self._released = True
            if self.attached:
                self.detach()
            self._cleanupResources()
            self.platformRelease()

    def getPid(self):
        """
        Return the pid for this Trace
        """
        return self.pid

    def getNormalizedLibNames(self):
        """
        Symbols are stored internally based off of
        "normalized" library names.  This method returns
        the list of normalized names for the loaded libraries.

        (probably only useful for writing symbol browsers...)
        """
        return list(self.getMeta("LibraryBases").keys())

    def getSymsForFile(self, libname):
        """
        Return the entire symbol list for the specified
        normalized library name.  The list is returned as
        "symtup" tuples of (va,size,name,type,fname).
        """
        self._loadBinaryNorm(libname)
        sym = self.getSymByName(libname)
        if sym is None:
            raise Exception('Invalid Library Name: %s' % libname)
        return sym.getSymList()

    def getNames(self):
        if not self.namecache:
            names = []
            for lib in self.getNormalizedLibNames():
                for sym in self.getSymsForFile(lib):
                    names.append((sym.value, str(sym)))
            self.namecache = sorted(names, key=lambda n: len(n[1]), reverse=True)
        return self.namecache

    def getSymByAddr(self, addr, exact=True):
        """
        Return an envi Symbol object for an address.
        Use exact=False to get the nearest previous match.
        """
        # NOTE: Override this from envi.SymbolResolver to do on-demand
        # file parsing.

        r = e_resolv.SymbolResolver.getSymByAddr(self, addr, exact=exact)
        if r is not None:
            return r

        # See if we need to parse the file.
        mmap = self.getMemoryMap(addr)
        if mmap is None:
            return None

        va, size, perms, fname = mmap

        if not self._loadBinary(fname):
            return None

        # Take a second shot after parsing
        return e_resolv.SymbolResolver.getSymByAddr(self, addr, exact=exact)

    def getSymByAddrThunkAware(self, va):
        '''
        TODO: DO NOT USE THIS FUNCTION, GOING AWAY.
        getBestSymEtc? depth / aggressiveness?

        for the given va:
        1. attempt to get the sym by using getSymByAddr
        2. if 1 fails, check the target of the branch for a sym.

        returns a tuple (sym, is_thunk).  sym is None if no sym is found.
        '''
        sym = self.getSymByAddr(va)
        if sym is not None:
            return str(sym), False

        try:
            op = self.parseOpcode(va)
            for tva, tflags in op.getTargets(emu=self):
                if tva is None:
                    continue

                sym = self.getSymByAddr(tva)
                if sym is not None:
                    return str(sym), True

        except Exception as e:
            # getTargets->readMemory error on bva
            logger.warning('getSymByAddrThunkAware: %s', e)

        return None, False

    def getSymByName(self, name):
        """
        Return an envi.Symbol object for the given name (or None)
        """
        self._loadBinaryNorm(name)
        return e_resolv.SymbolResolver.getSymByName(self, name)

    def setSymCachePath(self, path):
        '''
        Set the symbol cache path for the tracer.

        The "path" syntax is a ; seperated list of either directories
        or cobra URIs which implement the SymbolCache interface.

        Example:
            trace.setSymCachePath('/home/invisigoth/.envi/symcache;cobra://symbols.com/SymbolCache')

        NOTE: vdb automatically handles this with a config option
        '''
        self.symcache = e_symcache.SymbolCachePath(path)

    def searchSymbols(self, regex, libname=None):
        '''
        Search for symbols which match the given regular expression.  Specify
        libname as the "normalized" library name to only search the specified
        lib.

        Example:  for sym in trace.searchSymbols('.*CreateFile.*', 'kernel32'):
        '''
        reobj = re.compile(regex)
        if libname is not None:
            libs = [libname, ]
        else:
            libs = self.getNormalizedLibNames()

        ret = []
        for lname in libs:
            for sym in self.getSymsForFile(lname):
                symstr = str(sym)
                if reobj.match(symstr):
                    ret.append(sym)
        return ret

    def getRegisterContext(self, threadid=None):
        """
        Retrieve the envi.registers.RegisterContext object for the
        specified thread.  Use this API to iterate over threads
        register values without setting the global tracer thread context.
        """
        if threadid is None:
            threadid = self.getMeta("ThreadId")
        return self._cacheRegs(threadid)

#######################################################################
#
# We mirror the RegisterContext API using our own thread index based
# cache.  These APIs must stay in sync with envi.registers.RegisterContext
# NOTE: for now we only need to over-ride get/setRegister because all the
# higher level APIs call them.
#

    def getRegister(self, idx):
        ctx = self.getRegisterContext()
        return ctx.getRegister(idx)

    def setRegister(self, idx, value):
        ctx = self.getRegisterContext()
        ctx.setRegister(idx, value)

#######################################################################

    def allocateMemory(self, size, perms=e_mem.MM_RWX, suggestaddr=0):
        """
        Allocate a chunk of memory inside the target process' address
        space.  Memory wil be mapped rwx unless otherwise specified with
        perms=envi.memory.MM_FOO values. Optionally you may *suggest* an address
        to the allocator, but there is no guarentee.  Returns the mapped
        memory address.
        """
        self.requireNotRunning()
        self.mapcache = None # We may have a new memory map
        return self.platformAllocateMemory(size, perms=perms, suggestaddr=suggestaddr)

    def protectMemory(self, va, size, perms):
        """
        Change the page protections on the specified region of memory.
        See envi.memory for perms values.
        """
        self.requireNotRunning()
        self.mapcache = None # We may have new memory protections
        return self.platformProtectMemory(va, size, perms)

    def readMemory(self, address, size):
        """
        Read memory from address.  Areas that are NOT valid memory will be read
        back as \x00s (this probably goes in a mixin soon)
        """
        self.requireNotRunning()
        return self.platformReadMemory(int(address), int(size))

    def writeMemory(self, address, bytez):
        """
        Write the given bytes to the address in the current trace.
        """
        self.requireNotRunning()
        self.platformWriteMemory(int(address), bytez)

    def searchMemory(self, needle, regex=False):
        """
        Search all of process memory for a sequence of bytes.
        """
        ret = e_mem.IMemory.searchMemory(self, needle, regex=regex)
        self.setMeta('search', ret)
        self.setVariable('search', ret)
        return ret

    def searchMemoryRange(self, needle, address, size, regex=False):
        """
        Search a memory range for the specified sequence of bytes
        """
        ret = e_mem.IMemory.searchMemoryRange(self, needle, address, size, regex=regex)
        self.setMeta('search', ret)
        self.setVariable('search', ret)
        return ret

    def setMeta(self, name, value):
        """
        Set some metadata.  Metadata is a clean way for
        arbitrary trace consumers (and notifiers) to present
        and track additional information in trace objects.

        Any modules which use this *should* initialize them
        on attach (so when they get re-used they're clean)

        Some examples of metadata used:
        ShouldBreak - We're expecting a non-signal related break
        ExitCode - The int() exit code  (if exited)
        PendingSignal - The current signal

        """
        self.metadata[name] = value

    def getMeta(self, name, default=None):
        """
        Get some metadata.  Metadata is a clean way for
        arbitrary trace consumers (and notifiers) to present
        and track additional information in trace objects.

        If you specify a default and the key doesn't exist, not
        not only will the default be returned, but the key will
        be set to the default specified.
        """
        if default is not None:
            if name not in self.metadata:
                self.metadata[name] = default
        return self.metadata.get(name, None)

    def hasMeta(self, name):
        """
        Check to see if a metadata key exists... Mostly un-necessary
        as getMeta() with a default will set the key to the default
        if non-existant.
        """
        return name in self.metadata

    def getMode(self, name, default=False):
        """
        Get the value for a mode setting allowing
        for a clean default...
        """
        return self.modes.get(name, default)

    def setMode(self, name, value):
        """
        Set a mode setting...  This is ONLY valid
        if that mode has been iniitialized with
        initMode(name, value).  Otherwise, it's an
        unsupported mode for this platform ;)  cute huh?
        This way, platform sections can cleanly setmodes
        and such.
        """
        if name not in self.modes:
            raise Exception("Mode %s not supported on this platform" % name)
        self.modes[name] = bool(value)

    def injectso(self, filename):
        """
        Inject a shared object into the target of the trace.  So, on windows
        this is easy with InjectDll and on *nix... it's.. fugly...

        NOTE: This method will likely cause the trace to run.  Do not call from
              within a notifier!
        """
        self.requireNotRunning()
        self.platformInjectSo(filename)

    def ps(self):
        """
        Return a list of proccesses which are currently running on the
        system.
        (pid, name)
        """
        return self.platformPs()

    def addBreakByExpr(self, symname, fastbreak=False):
        '''
        Add a breakpoint by resolving an expression.  This will create
        the Breakpoint object for you and add it to the trace.  It
        returns the newly created breakpoint id.

        Optionally, set fastbreak=True to have the breakpoint behave in
        "fast break" mode which automatically continues execution and does
        not fire notifiers for the breakpoint.

        Example: trace.addBreakByExpr('kernel32.CreateFileA + ecx')
        '''
        bp = Breakpoint(None, expression=symname)
        bp.fastbreak = fastbreak
        return self.addBreakpoint(bp)

    def addBreakByAddr(self, va, fastbreak=False):
        '''
        Add a breakpoint by address.  This will create the Breakpoint
        object for you and add it to the trace.  It returns the newly
        created breakpoint id.

        Optionally, set fastbreak=True to have the breakpoint behave in
        "fast break" mode which automatically continues execution and does
        not fire notifiers for the breakpoint.

        Example: trace.addBreakByAddr(0x7c770308)
        '''
        bp = Breakpoint(va)
        bp.fastbreak = fastbreak
        return self.addBreakpoint(bp)

    def addBreakpoint(self, breakpoint):
        """
        Add a breakpoint/watchpoint to the trace.  The "breakpoint" argument
        is a vtrace Breakpoint/Watchpoint object or something that extends it.

        To add a basic breakpoint use:
        trace.addBreakpoint(vtrace.Breakpoint(address))

        NOTE: expression breakpoints do *not* get evaluated in fastbreak mode

        This will return the internal ID given to the new breakpoint
        """
        breakpoint.inittrace(self)

        breakpoint.id = self.nextBpId()
        addr = breakpoint.resolveAddress(self)

        if addr is None:
            self.bpbyid[breakpoint.id] = breakpoint
            self.deferred.append(breakpoint)
            return breakpoint.id

        if addr in self.breakpoints:
            raise Exception("ERROR: Duplicate break for address 0x%.8x" % addr)

        self.bpbyid[breakpoint.id] = breakpoint
        self.breakpoints[addr] = breakpoint

        # fastbreaks are always active... (except when they're not...)
        if breakpoint.fastbreak:
            breakpoint.activate(self)

        return breakpoint.id

    def removeBreakpoint(self, id):
        """
        Remove the breakpoint with the specified ID
        """
        self.requireAttached()
        bp = self.bpbyid.pop(id, None)
        if bp is not None and not bp.stealthbreak:
            bp.deactivate(self)
            if bp in self.deferred:
                self.deferred.remove(bp)
            else:
                self.breakpoints.pop(bp.address, None)

            # If the bp is also curbp, set curbp to None
            if self.curbp == bp:
                self.curbp = None

        # Remove cached breakpoint code
        Breakpoint.bpcodeobj.pop(id, None)

    def _updateBreakAddresses(self):
        """
        Update breakpoint address resolution (if unresolved).
        Intended to be run after events which change the namespace, such as
        NOTIFY_LOAD_LIBRARY events
        """
        for bp in self.deferred:
            bp.resolveAddress(self)
            if bp.address is not None:
                self.breakpoints[bp.address] = bp
                self.deferred.remove(bp)
                bp.activate(self)
                logger.warning("Resolved bp address: %r", bp)

    def getCurrentBreakpoint(self):
        """
        Return the current breakpoint otherwise None
        """
        return self.curbp

    def getBreakpoint(self, id):
        """
        Return a reference to the breakpoint with the requested ID.

        NOTE: NEVER set locals or use things like setBreakpointCode()
        method on return'd breakpoint objects as they may be remote
        and would then be *coppies* of the bp objects. (use the trace's
        setBreakpointCode() instead).
        """
        return self.bpbyid.get(id)

    def getBreakpointByAddr(self, va):
        '''
        Return the breakpoint object (or None) for a given virtual address.
        '''
        return self.breakpoints.get(va)

    def getBreakpoints(self):
        """
        Return a list of the current breakpoints.
        """
        return list(self.bpbyid.values())

    def getBreakpointEnabled(self, bpid):
        """
        An accessor method for returning if a breakpoint is
        currently enabled.
        NOTE: code which wants to be remote-safe should use this
        """
        bp = self.getBreakpoint(bpid)
        if bp is None:
            raise Exception("Breakpoint %d Not Found" % bpid)
        return bp.isEnabled()

    def setBreakpointEnabled(self, bpid, enabled=True):
        """
        An accessor method for setting a breakpoint enabled/disabled.

        NOTE: code which wants to be remote-safe should use this
        """
        bp = self.getBreakpoint(bpid)
        if bp is None or bp.stealthbreak:
            raise Exception("Breakpoint %d Not Found" % bpid)
        if not enabled: # To catch the "disable" of fastbreaks...
            bp.deactivate(self)
        return bp.setEnabled(enabled)

    def setBreakpointCode(self, bpid, pystr):
        """
        Because breakpoints are potentially on the remote debugger
        and code is not pickleable in python, special access methods
        which takes strings of python code are necessary for the
        vdb interface to quick script breakpoint code.  Use this method
        to set the python code for this breakpoint.
        """
        bp = self.getBreakpoint(bpid)
        if bp is None:
            raise Exception("Breakpoint %d Not Found" % bpid)
        bp.setBreakpointCode(pystr)

    def getBreakpointCode(self, bpid):
        """
        Return the python string of user specified code that will run
        when this breakpoint is hit.
        """
        bp = self.getBreakpoint(bpid)
        if bp is not None:
            return bp.getBreakpointCode()
        return None

    def call(self, address, args, convention=None):
        """
        Setup the "stack" and call the target address with the following
        arguments.  If the argument is a string or a buffer, copy that into
        memory and hand in the argument.

        The current state of ALL registers are returned as a dictionary at the
        end of the call...

        Additionally, a "convention" string may be specified that the underlying
        platform may be able to interpret...
        """
        self.requireNotRunning()
        return self.platformCall(address, args, convention)

    def registerNotifier(self, event, notifier):
        """
        Register a notifier who will be called for various
        events.  See NOTIFY_* constants for handler hooks.
        """
        nlist = self.notifiers.get(event, None)
        if nlist:
            nlist.append(notifier)
        else:
            nlist = []
            nlist.append(notifier)
            self.notifiers[event] = nlist

    def deregisterNotifier(self, event, notifier):
        nlist = self.notifiers.get(event, [])
        if notifier in nlist:
            nlist.remove(notifier)

    def getNotifiers(self, event):
        return self.notifiers.get(event, [])

    def requireNotExited(self):
        '''
        Call in a method that requires the trace to have not exited.
        '''
        if self.exited:
            raise Exception('ERROR - Request invalid for trace which exited')

    def requireNotRunning(self):
        '''
        Call in a method that requires the debugger the be attached and not
        running.
        '''
        self.requireAttached()
        if self.isRunning():
            raise Exception('ERROR - trace is running; use "break" before running the specified command')

    def requireAttached(self):
        '''
        Call in a method that requires the debugger to be attached.
        '''
        if not self.attached:
            raise Exception('ERROR - attach to a process first')

    def getFds(self):
        """
        Get a list of (fd, type, bestname) pairs.  This is MOSTLY useful
        for HUMON consumtion...  or giving HUMONs consumption...
        """
        self.requireNotRunning()
        if not self.fds:
            self.fds = self.platformGetFds()
        return self.fds

    def getMemoryMaps(self):
        """
        Return a list of the currently mapped memory for the target
        process.  This is acomplished by calling the platform's
        platformGetMaps() mixin method.  This will also cache the
        results until CONTINUE.  The format is (addr, len, perms, file).
        """
        self.requireNotRunning()
        if not self.mapcache:
            self.mapcache = self.platformGetMaps()
        return self.mapcache

    def getMemoryFault(self):
        '''
        If the most receent event is a memory access error, this API will
        return a tuple of (<addr>, <perm>) on supported platforms.  Otherwise,
        a (None, None) will result.

        Example:
        import envi.memory as e_mem
        vaddr, vperm = trace.getMemoryFault()
        if vaddr is not None:
            print('Memory Fault At: 0x%.8x (perm: %d)' % (vaddr, vperm))
        '''
        return self.platformGetMemFault()

    def isAttached(self):
        '''
        Return true or false if this trace's target processing is attached.
        '''
        return self.attached

    def isRunning(self):
        '''
        Return true or false if this trace's target process is running.
        '''
        return self.running

    def hasExited(self):
        '''
        Return true or false if this trace's target process has exited.
        '''
        return self.exited

    def isRemote(self):
        '''
        Return true or false if this trace's target process is a CobraProxy
        object to a trace on another system.
        '''
        return False

    def enableAutoContinue(self, event):
        """
        Put the tracer object in to AutoContinue mode
        for the specified event.  To make all events
        continue running see RunForever mode in setMode().
        """
        if event not in self.auto_continue:
            self.auto_continue.append(event)

    def disableAutoContinue(self, event):
        """
        Disable Auto Continue for the specified
        event.
        """
        if event in self.auto_continue:
            self.auto_continue.remove(event)

    def getAutoContinueList(self):
        """
        Retrieve the list of vtrace notification events
        that will be auto-continued.
        """
        return list(self.auto_continue)

    def parseExpression(self, expression):
        """
        Parse a python expression with many useful helpers mapped
        into the execution namespace.

        Example: trace.parseExpression("ispoi(ecx+ntdll.RtlAllocateHeap)")
        """
        vlocs = VtraceExpressionLocals(self)
        return int(e_expr.evaluate(expression, vlocs))

    def sendBreak(self):
        """
        Send an asynchronous break signal to the target process.
        This is only valid if the target is actually running...
        """
        self.requireAttached()
        self.setMode("RunForever", False)
        self.setMeta("ShouldBreak", True)
        self.platformSendBreak()
        time.sleep(0.01)
        # If we're non-blocking, we gotta wait...
        if self.getMode("NonBlocking", True):
            while self.isRunning():
                time.sleep(0.01)

    def getStackTrace(self):
        """
        Returns a list of (instruction pointer, stack frame) tuples.
        If stack tracing results in an error, the error entry will
        be (-1, -1).  Otherwise most platforms end up with 0, 0 as
        the top stack frame
        """
        # FIXME thread id argument!
        return self.archGetStackTrace()

    def getThreads(self):
        """
        Get a dictionary of <threadid>:<tinfo> pairs where
        tinfo is platform dependant, but is typically either
        the top of the stack for that thread, or the TEB on
        win32
        """
        if not self.threadcache:
            self.threadcache = self.platformGetThreads()
        return self.threadcache

    def getCurrentThread(self):
        '''
        Return the thread id of the currently selected thread.
        '''
        return self.getMeta('ThreadId')

    def selectThread(self, threadid):
        """
        Set the "current thread" context to the given thread id.
        (For example stack traces and register values will depend
        on the current thread context).  By default the thread
        responsible for an "interesting event" is selected.
        """
        if threadid not in self.getThreads():
            raise Exception("ERROR: Invalid threadid chosen: %d" % threadid)
        self.requireNotRunning()
        self.platformSelectThread(threadid)
        self.setMeta("ThreadId", threadid)

    def isThreadSuspended(self, threadid):
        """
        Used to determine if a thread is suspended.
        """
        return self.sus_threads.get(threadid, False)

    def suspendThread(self, threadid):
        """
        Suspend a thread by ID.  This will mean that on continuing
        the trace, the suspended thread will not be scheduled.
        """
        self.requireNotRunning()
        if self.sus_threads.get(threadid):
            raise Exception("The specified thread is already suspended")
        if threadid not in self.getThreads().keys():
            raise Exception("There is no thread %d!" % threadid)
        self.platformSuspendThread(threadid)
        self.sus_threads[threadid] = True

    def resumeThread(self, threadid):
        """
        Resume a suspended thread.
        """
        self.requireNotRunning()
        if not self.sus_threads.get(threadid):
            raise Exception("The specified thread is not suspended")
        self.platformResumeThread(threadid)
        self.sus_threads.pop(threadid)

    def injectThread(self, pc):
        """
        Create a new thread inside the target process.  This thread
        will begin execution on the next process run().
        """
        self.requireNotRunning()
        #self.platformInjectThread(pc)
        pass

    def joinThread(self, threadid):
        '''
        Run the trace in a loop until the specified thread exits.
        '''
        self.setMode('RunForever', True)
        self._join_thread = threadid
        # Temporarily make run/wait blocking
        nb = self.getMode('NonBlocking')
        self.setMode('NonBlocking', False)
        self.run()
        self.setMode('NonBlocking', nb)

    def getStructNames(self, namespace=None):
        '''
        This method returns either the structure names, or
        the structure namespaces that the target tracer is aware
        of.  If "namespace" is specified, it is structures within
        that namespace, otherwise it is "known namespaces"

        Example: namespaces = trace.getStructNames()
                 ntdll_structs = trace.getStructNames(namespace='ntdll')
        '''
        if namespace:
            return self.vsbuilder.getVStructNames(namespace=namespace)
        return self.vsbuilder.getVStructNamespaceNames()

    def getStruct(self, sname, va=None):
        """
        Retrieve a vstruct structure optionally populated with memory from
        the specified address.  Returns a standard vstruct object.
        """
        # Check if we need to parse symbols for a library
        libbase = sname.split('.')[0]
        self._loadBinaryNorm(libbase)

        if self.vsbuilder.hasVStructNamespace(libbase):
            vs = self.vsbuilder.buildVStruct(sname)

        # FIXME this is deprecated and should die...
        else:
            vs = vstruct.getStructure(sname)

        if vs is None:
            return None

        if va is None:
            return vs

        bytez = self.readMemory(va, len(vs))
        vs.vsParse(bytez)
        return vs

    def setVariable(self, name, value):
        """
        Set a named variable in the trace which may be used in
        subsequent VtraceExpressions.

        Example:
        trace.setVariable("whereiam", trace.getProgramCounter())
        """
        self.localvars[name] = value

    def getVariable(self, name):
        """
        Get the value of a previously set variable name.
        (or None on not found)
        """
        return self.localvars.get(name)

    def getVariables(self):
        """
        Get the dictionary of named variables.
        """
        return dict(self.localvars)

    def hex(self, value):
        """
        Much like the python hex routine, except this will automatically
        pad the value's string length out to pointer width.
        """
        width = self.arch.getPointerSize()
        return e_bits.hex(value, width)

    def vprint(self, msg, addnl=True):
        if addnl:
            msg = msg + "\n"
        return print(msg)

    def buildNewTrace(self, **kwargs):
        '''
        Build a new/clean trace "like" this one.  For platforms where a
        special trace was handed in, this allows initialization of a new one.
        For most implementations, this is very simple....

        Example:
            if need_another_trace:
                newt = trace.buildNewTrace()
        '''
        return self.__class__(**kwargs)

class TraceGroup(Notifier, v_util.TraceManager):
    """
    Encapsulate several traces, run them, and continue to
    handle their event notifications.
    """
    def __init__(self):
        Notifier.__init__(self)
        v_util.TraceManager.__init__(self)
        self.traces = {}
        self.go = True # A little ghetto switch for those who read the source

        # We are a notify all notifier by default
        self.registerNotifier(NOTIFY_ALL, self)

        self.setMode("NonBlocking", True)

    def setMeta(self, name, value):
        """
        A trace group's setMeta function will set "persistant" metadata
        which will be added again to any trace on attach.  Additionally,
        setting metadata on a tracegroup will cause all current traces
        to get the update as well....
        """
        v_util.TraceManager.setMeta(self, name, value)
        for trace in self.traces.values():
            trace.setMeta(name, value)

    def setMode(self, name, value):
        v_util.TraceManager.setMode(self, name, value)
        for trace in self.getTraces():
            trace.setMode(name, value)

    def detachAll(self):
        """
        Detach from ALL the currently targetd processes
        """
        for trace in self.traces.values():
            try:
                if trace.isRunning():
                    trace.sendBreak()
                trace.detach()
            except:
                pass

    def run(self):
        """
        Our run method  is a little different than a traditional
        trace. It will *never* block.
        """
        if len(self.traces.keys()) == 0:
            raise Exception("ERROR - can't run() with no traces!")

        for trace in self.traces.values():

            if trace.exited:
                self.traces.pop(trace.pid)
                trace.detach()
                continue

            if not trace.isRunning():
                trace.run()

    def execTrace(self, cmdline):
        trace = getTrace()
        self._initTrace(trace)
        trace.execute(cmdline)
        self.traces[trace.getPid()] = trace
        return trace

    def addTrace(self, proc):
        """
        Add a new tracer to this group the "proc" argument
        may be either an int for a pid (which we will attach
        to) or an already attached (and broken) tracer object.
        """
        if isinstance(proc, int):
            trace = getTrace()
            self._initTrace(trace)
            self.traces[proc] = trace
            try:
                trace.attach(proc)
            except:
                self.delTrace(proc)
                raise

        else:  # Hopefully a tracer object... if not.. you're dumb.
            trace = proc
            self._initTrace(trace)
            self.traces[trace.getPid()] = trace

        return trace

    def getTrace(self):
        '''
        Similar to vtrace.getTrace(), but also init's
        the trace for being managed by a TraceGroup.

        Example:
            tg = TraceGroup()
            t = tg.getTrace()
            t....
        '''
        t = getTrace()
        self.addTrace(t)
        return t

    def _initTrace(self, trace):
        """
         - INTERNAL -
        Setup a tracer object to be ready for being in this
        trace group (setup modes and notifiers).  Only addTrace()
        and execTrace() probably need to be aware of this.
        """
        self.manageTrace(trace)

    def delTrace(self, pid):
        """
        Remove a trace from the current TraceGroup
        """
        trace = self.traces.pop(pid, None)
        self.unManageTrace(trace)

    def getTraces(self):
        """
        Return a list of the current traces
        """
        return list(self.traces.values())

    def getTraceByPid(self, pid):
        """
        Return the the trace for process PID if we're
        already attached.  Return None if not.
        """
        return self.traces.get(pid, None)

    def notify(self, event, trace):
        # Remove this trace, and free it
        # on the server if present
        if event == NOTIFY_EXIT:
            self.delTrace(trace.getPid())

class VtraceExpressionLocals(e_expr.MemoryExpressionLocals):
    """
    A class which serves as the namespace dictionary during the
    evaluation of an expression on a tracer.
    """
    def __init__(self, trace):
        e_expr.MemoryExpressionLocals.__init__(self, trace, symobj=trace)
        self.trace = trace
        self.update({
            'trace': trace,
            'vtrace': vtrace
        })
        self.update({
            'frame': self.frame,
            'teb': self.teb,
            'bp': self.bp,
            'meta': self.meta,
            'go': self.go,
        })

    def __getitem__(self, name):

        # Check registers
        if self.trace.isAttached() and not self.trace.isRunning():

            regs = self.trace.getRegisters()
            r = regs.get(name, None)
            if r is not None:
                return r

        # Check local variables
        locs = self.trace.getVariables()
        r = locs.get(name, None)
        if r is not None:
            return r

        # Check the loaded libraries
        for lib in self.trace.getNormalizedLibNames():
            for sym in self.trace.getSymsForFile(lib):
                if str(sym) == name:
                    return sym

        return e_expr.MemoryExpressionLocals.__getitem__(self, name)

    def go(self):
        '''
        A shortcut for trace.runAgain() which may be used in
        breakpoint code (or similar even processors) to begin
        execution again after event processing...
        '''
        self.trace.runAgain()

    def frame(self, index):
        """
        Return the address of the saved base pointer for
        the specified frame.

        Usage: frame(<index>)
        """
        stack = self.trace.getStackTrace()
        return stack[index][1]

    def teb(self, threadnum=None):
        """
        The expression teb(threadid) will return whatever the
        platform stores as the int for threadid.  In the case
        of windows, this is the TEB, others may be the thread
        stack base or whatever.  If threadid is left out, it
        uses the threadid of the current thread context.
        """
        if threadnum is None:
            # Get the thread ID of the current Thread Context
            threadnum = self.trace.getMeta("ThreadId")

        teb = self.trace.getThreads().get(threadnum, None)
        if teb is None:
            raise Exception("ERROR - Unknown Thread Id %d" % threadnum)

        return teb

    def bp(self, bpid):
        """
        The expression bp(0) returns the resolved address of the given
        breakpoint
        """
        bp = self.trace.getBreakpoint(bpid)
        if bp is None:
            raise Exception("Unknown Breakpoint ID: %d" % bpid)
        return bp.resolveAddress(self.trace)

    def meta(self, name):
        """
        An expression friendly (terse) way to get trace metadata
        (equiv to trace.getMeta(name))

        Example: meta("foo")
        """
        return self.trace.getMeta(name)

def reqTargOpt(opts, targ, opt, valstr='<value>'):
    val = opts.get( opt )
    if val is None:
        raise Exception('Target "%s" requires option: %s=%s' % (targ, opt, valstr))
    return val

def getTrace(target=None, **kwargs):
    """
    Return a tracer object appropriate for this platform.
    This is the function you will use to get a tracer object
    with the appropriate ancestry for your host.

    ex. mytrace = vtrace.getTrace()


    NOTE: Use the release() method on the tracer once debugging
          is complete.  This releases the tracer thread and allows
          garbage collection to function correctly.

    Some specialized tracers may be constructed by specifying the "target"
    name from one of the following list.  Additionally, each "specialized"
    tracer may require additional kwargs (which are listed).


    Examples:
        # A tracer for *this* os
        t = vtrace.getTrace()

        # A tracer for the gdbstub debugging a vmware 32bit hypervisor
        t = vtrace.getTrace(target='vmware32', host='localhost', port=8832)

    Targets:

    Alpha Targets:

    vmware32    -
        host=<host>     ( probably 'localhost' )
        port=<port>     ( probably 8832 )

    """
    if target == 'gdbserver':

        host = reqTargOpt(kwargs, 'gdbserver', 'host', '<host>')
        port = reqTargOpt(kwargs, 'gdbserver', 'port', '<port>')
        arch = reqTargOpt(kwargs, 'gdbserver', 'arch', '<i386|amd64|arm>')
        plat = reqTargOpt(kwargs, 'gdbserver', 'plat', '<windows|linux>')

        if arch not in ('i386', 'amd64', 'arm'):
            raise Exception('Invalid arch specified for "gdbserver" target: %s' % arch)

        if plat not in ('windows', 'linux'):
            raise Exception('Invalid plat specified for "gdbserver" target: %s' % plat)

    if target == 'vmware32':

        import vtrace.platforms.vmware as vt_vmware

        host = reqTargOpt(kwargs, 'vmware32', 'host', '<host>')
        port = int( reqTargOpt(kwargs, 'vmware32', 'port', '<port>') )

        plat = 'windows'

        #plat = reqTargOpt(kwargs, 'vmware32', 'plat', '<windows|linux>')
        #if plat not in ('windows', 'linux'):
            #raise Exception('Invalid plat specified for "vmware32" target: %s' % plat)

        return vt_vmware.VMWare32WindowsTrace( host=host, port=port )

    if remote: #We have a remote server!
        return getRemoteTrace()

    # From here down, we're trying to build a trace for *this* platform!

    os_name = platform.system().lower() # Like "linux", "darwin","windows"
    arch = envi.getCurrentArch()

    if os_name == "linux":
        import vtrace.platforms.linux as v_linux
        if arch == "amd64":
            return v_linux.LinuxAmd64Trace()

        elif arch == "i386":
            return v_linux.Linuxi386Trace()

        # Keep separate just in case
        elif arch == "armv7l":
            return v_linux.LinuxArmTrace()

        elif arch == "armv6l":
            return v_linux.LinuxArmTrace()

        else:
            raise Exception("Sorry, no linux support for %s" % arch)

    elif os_name == "freebsd":

        import vtrace.platforms.freebsd as v_freebsd

        if arch == "i386":
            return v_freebsd.FreeBSDi386Trace()

        elif arch == "amd64":
            return v_freebsd.FreeBSDAmd64Trace()

        else:
            raise Exception("Sorry, no FreeBSD support for %s" % arch)

    elif os_name == "sunos5":
        raise Exception("Solaris needs porting!")
        #import vtrace.platforms.posix as v_posix
        #import vtrace.platforms.solaris as v_solaris
        #ilist.append(v_posix.PosixMixin)
        #if arch == "i386":
            #import vtrace.archs.intel as v_intel
            #ilist.append(v_intel.i386Mixin)
            #ilist.append(v_solaris.SolarisMixin)
            #ilist.append(v_solaris.Solarisi386Mixin)

    elif os_name == "darwin":

        #if 9 not in os.getgroups():
            #print 'You MUST be in the procmod group....'
            #print 'Use: sudo dscl . append /Groups/procmod GroupMembership invisigoth'
            #print '(put your username in there unless you want to put me in too... ;)'
            #raise Exception('procmod group membership required')
        if os.getuid() != 0:
            logger.error('For NOW you *must* be root.  There are some crazy MACH perms...')
            raise Exception('You must be root for now (on OSX)....')

        logger.warning('Also... the darwin port is not even REMOTELY working yet.  Solid progress though...')

        #'sudo dscl . append /Groups/procmod GroupMembership invisigoth'
        #'sudo dscl . read /Groups/procmod GroupMembership'
        import vtrace.platforms.darwin as v_darwin
        if arch == 'i386':
            return v_darwin.Darwini386Trace()
        elif arch == 'amd64':
            return v_darwin.DarwinAmd64Trace()
        else:
            raise Exception('Darwin not supported on %s (only i386...)' % arch)

    elif os_name in ['microsoft', 'windows']:

        import vtrace.platforms.win32 as v_win32

        if arch == "i386":
            return v_win32.Windowsi386Trace()

        elif arch == "amd64":
            return v_win32.WindowsAmd64Trace()

        else:
            raise Exception("Windows with arch %s is not supported!" % arch)

    else:

        raise Exception("ERROR - OS %s not supported yet" % os_name)

def interact(pid=0, server=None, trace=None):

    """
    Just a cute and dirty way to get a tracer attached to a pid
    and get a python interpreter instance out of it.
    """

    global remote
    remote = server

    if trace is None:
        trace = getTrace()
        if pid:
            trace.attach(pid)

    mylocals = {}
    mylocals["trace"] = trace

    code.interact(local=mylocals)

def getEmu(trace, arch=envi.ARCH_DEFAULT):
    '''
    See comment for emulator from trace (in envitools); does not set any
    registers or mem.

    TODO: this really belongs in envitools, or somewhere else, but putting it
    in envitools causes a circular import problem due to the TraceEmulator.
    '''
    if arch == envi.ARCH_DEFAULT:
        arch_name = trace.getMeta('Architecture')
    else:
        arch_name = envi.getArchById(trace.arch)

    arch_mod = envi.getArchModule(arch_name)
    emu = arch_mod.getEmulator()
    return emu



import logging
import unittest

import vivisect.codegraph as v_codegraph
import vivisect.tests.helpers as helpers

logger = logging.getLogger(__name__)

class GraphCoreTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.vw = helpers.getTestWorkspace('windows', 'amd64', 'firefox.exe')

    def test_vivisect_codegraph_firefox(self):
        fva = 0x1400041a0
        cg = v_codegraph.FuncBlockGraph(self.vw, fva)
        # TODO: add 0x140020920 as a test once the jump table stuff gets fixed
        # LauncherMain in the firefox source code:
        cg.addEntryPoint(fva)
        self.assertTrue(cg.isCodeBlockNode(fva))
        self.assertFalse(cg.isCodeBlockNode(fva + 1))
        cb = cg.getCodeBlockNode(fva)
        self.assertEqual(cg.getCodeBlockBounds(cb), (fva, 110))
        self.assertIsNotNone(cg.getNodeByVa(0x140004213))
        self.assertEqual(len(cg.nodevas), self.vw.getFunctionMeta(fva, 'InstructionCount'))

        # LaunchUnelevated.cpp, GetMediumIntegrityToken
        fva = 0x140004020
        cg = v_codegraph.FuncBlockGraph(self.vw, fva)
        self.assertEqual(len(cg.nodevas), self.vw.getFunctionMeta(fva, 'InstructionCount'))
        self.assertEqual(len(cg.nodes), 11)
        self.assertEqual(len(cg.nodes), self.vw.getFunctionMeta(fva, 'BlockCount'))
        codeblocks = [
            0x140004020,
            0x1400040f3,
            0x14000406d,
            0x14000413d,
            0x14000409c,
            0x14000416e,
            0x1400040d4,
            0x1400040db,
            0x1400040de,
            0x1400040e8,
            0x14000411f,
        ]
        for c in codeblocks:
            self.assertTrue(cg.isCodeBlockNode(c))



'''
Late-PLT analysis.

During codeflow analysis, we detect PLT entries organically (and consequently
*accurately*), but often not all of them.  This analysis module applies a few
different algorithms to identify the ones which were not discovered during
autoanalysis.

Formerly, we attempted to do this analysis before all other code analysis
but due to so many varieties of PLT creation by different compilers on
numerous hardware platforms and OSes, that method lead to many errors in PLT-
identification, often hampering vital parts of vulnerability-research tools.
We finally pulled the plug on preemptive analysis, in favor of letting auto-
analysis identify "truth" and then filling in the gaps.
'''


import logging
import vivisect
import envi.common as e_cmn

from . import elfplt

logger = logging.getLogger(__name__)

def analyze(vw):
    logger.info('ELFPLT late-analysis')

    for pltva, pltsz in elfplt.getPLTs(vw):
        try:
            analyzePLT(vw, pltva, pltsz)

        except Exception as e:
            logger.warning("Error in PLT-late analysis: %r", e, exc_info=True)


def analyzePLT(vw, pltva, pltsz):
    '''
    Analyze a specific PLT section (there are often more than one and they often
    different in format)

    We make use of two different algorithms for each.
    The first algorithm measures the distance from the start of known good PLT
    functions and the GOT-referencing branch.  The only weakness is that this
    method requires that the xref to the GOT is already identified (without
    throwing and emulator in there).

    The second algorithm measures the distance between known good PLT entries
    and then attempts to identify divisors (up to 16 splits) which would make
    more than one PLT function fit between them.  These attempted splits are
    validated using heuristics of the potential functions which would be created
    by the division.  PLT entries in the same PLT section are incredibly similar
    (not including the LazyLoader sometimes found at the beginning of a PLT)
    '''
    logger.info("PLT Section -  Address: 0x%x  Size: %d", pltva, pltsz)
    gotva = gotsz = None
    fname = vw.getFileByVa(pltva)

    # find functions currently defined in this PLT
    curpltset = set()
    for fva in vw.getFunctions():
        if pltva <= fva < (pltva+pltsz) and fva not in curpltset:
            logger.debug("existing PLT Function: 0x%x", fva)
            curpltset.add(fva)

    curplts = list(curpltset)
    curplts.sort()

    if len(curplts):    # make sure we have *any* functions
        # just make the first thing a function.  this should take care of any LazyLoader
        # or at the very least, the first PLT entry.  it should also keep us from backing
        # up too far during PLT-Func-Distance algorithm.
        logger.debug('analyzePLT(0x%x, 0x%x): making first location the PLT into a function', pltva, pltsz)
        vw.makeFunction(pltva)

    # now figure out the distance from function start to the GOT xref 
    # and between PLT functions
    lastva = pltva
    jmpheur = {}
    distanceheur = {}
    for fva in curplts:
        fsz = vw.getFunctionMeta(fva, 'Size')
        gotva, gotsz = elfplt.getGOTByFilename(vw, fname)

        offset = 0
        while offset < fsz:
            locva, lsz, ltype, ltinfo = vw.getLocation(fva + offset)
            xrefsfrom = vw.getXrefsFrom(locva)
            for xrfr, xrto, xrtype, xrtinfo in xrefsfrom:
                if gotva <= xrto < gotva+gotsz:
                    offcnt = jmpheur.get(offset, 0)
                    jmpheur[offset] = offcnt + 1

            offset += lsz

        # capture distance between functions
        delta = fva - lastva
        distanceheur[delta] = distanceheur.get(delta, 0) + 1
        lastva = fva

    if None not in (gotva, gotsz):
        logger.debug("GOT/size: 0x%x/0x%x", gotva, gotsz)

    # GOT-XREF-Offset method
    if len(jmpheur):
        fillPLTviaGOTXrefs(vw, jmpheur, pltva, pltsz)

    else:
        logger.info("analyzePLT(0x%x, 0x%x) skipping GOT-XREF-Offset method: no existing functions found", pltva, pltsz)

    # PLT-Func-Distance method
    if len(distanceheur):
        fillPLTGaps(vw, curplts, distanceheur, pltva, pltsz)

    else:
        logger.info("skipping analyzePLT(0x%x, 0x%x) (PLT-Func-Distance method): no existing functions found", pltva, pltsz)

    logger.info("elfplt_late (done): pltva: 0x%x, %d", pltva, pltsz)


def fillPLTviaGOTXrefs(vw, jmpheur, pltva, pltsz):
    '''
    This PLT-placement algorithm measures the distance from the start of known good
    PLT functions and the GOT-referencing branch.  The only weakness is that this
    method requires that the xref to the GOT is already identified (without
    throwing and emulator in there).

    This method seems to work for many PLT's, but only if the xref is identifiable
    without function analysis.  it fails on i386-pic code which uses ebx  (which
    is handed into the call as a base address)
    '''
    # if we have what we need... scroll through all the non-functioned area
    # looking for GOT-xrefs
    offbycnt = [(cnt, off) for off, cnt in jmpheur.items()]
    offbycnt.sort(reverse=True)
    logger.debug("PLT Segment Data: %r", vw.getSegment(pltva))
    cnt, realoff = offbycnt[0]

    # now roll through the PLT space and look for GOT-references from 
    # locations that aren't in a function
    logger.info("Scanning for Xrefs into the GOT to determine PLT function starts")
    offset = 0
    while offset < pltsz:
        locva, lsz, ltype, ltinfo = vw.getLocation(pltva + offset)

        xrefsfrom = vw.getXrefsFrom(locva)
        toGOT = False
        for xrfr, xrto, xrtype, xrtinfo in xrefsfrom:
            if isGOT(vw, xrto):
                # make sure we're pointing at a valid import
                toloc = vw.getLocation(xrto)
                if toloc is None:
                    continue

                tlva, tlsz, tltype, tlinfo = toloc
                if tltype not in (vivisect.LOC_POINTER, vivisect.LOC_IMPORT):
                    continue

                toGOT = True

        logger.debug("loc: 0x%x   xrefs: %r (toGOT: %r)", locva, xrefsfrom, toGOT)
        if toGOT:
            # we have an xref into the GOT and no function.  go!
            funcstartva = locva - realoff
            if vw.getFunction(locva) != funcstartva:
                logger.debug("New PLT Function: 0x%x (GOT jmp: 0x%x)", funcstartva, locva)

                # if our intended location is not currently part of a PLT function, make it one
                curfuncva = vw.getFunction(funcstartva)
                if curfuncva is None or not isPLT(vw, curfuncva):
                    vw.makeFunction(funcstartva)
                else:
                    logger.debug("attempting to make function at 0x%x, which is already a member of 0x%x",
                            funcstartva, vw.getFunction(funcstartva))

        offset += lsz


def fillPLTGaps(vw, curplts, distanceheur, pltva, pltsz):
    '''
    This PLT-placement algorithm measures the distance between known good PLT entries
    and then attempts to identify divisors (up to 16 splits) which would make
    more than one PLT function fit between them.  These attempted splits are
    validated using heuristics of the potential functions which would be created
    by the division.  PLT entries in the same PLT section are incredibly similar
    (not including the LazyLoader sometimes found at the beginning of a PLT)
    '''
    ######## let's attempt to identify the smallest common distance between functions
    # what's the smallest distance between functions that
    logger.info("Using known good PLT entries to fill in the rest of the PLT (based on gaps)")
    curpltcnt = len(curplts)
    minimumbar = 1 + (curpltcnt // 100)

    # cull the herd.  our winner wants to be at least greater than one entry
    for delta, count in list(distanceheur.items()):
        if count < minimumbar:
            distanceheur.pop(delta)

        elif delta == 0:  # can't have 0.  skip the first PLT function
            distanceheur.pop(delta)

    workdist = list(distanceheur.items())
    workdist.sort()

    if not len(workdist):
        logger.info('... bailing, no existing PLT functions in this section to compare to...')

    else:
        # first entry should be our guy...
        funcdist = workdist[0][0]
        logger.debug("funcdist: 0x%x", funcdist)

        idx = getGoodIndex(curplts, funcdist)
        if not idx:
            return
        fva = curplts[idx]

        ### magic check:  we are working off incomplete data... it's completely possible our
        # gap measurement is twice as big as it should be (or 3 times).  how to check is 
        # where the magic comes in...
        funcsz = vw.getFunctionMeta(fva, 'Size')
        for divisor in range(funcdist, 0, -1):
            logger.log(e_cmn.SHITE, 'attempting to divide funcdist (0x%x) by %d', funcdist, divisor)

            # does the gap between functions support splitting?
            if funcdist < (divisor * funcsz):
                logger.log(e_cmn.SHITE, '.. not big enough')
                continue

            newdist = funcdist // divisor
            # does the funcdist split evenly?
            if newdist * divisor != funcdist:
                logger.log(e_cmn.SHITE, ".. doesn't divide evenly(newdist: 0x%x  divisor: 0x%x   funcdist: 0x%x)", newdist, divisor, funcdist)
                continue

            # do the opcodes support this size split?
            if not compareFuncs(vw, fva, fva + newdist, funcsz):
                logger.log(e_cmn.SHITE, ".. functions don't match!")
                continue

            break

        # should end up dividing by 1 if not divisible
        if divisor > 1:
            logger.info("dividing our lowest function (0x%x) distance by %d", funcdist, divisor)
            funcdist //= divisor

        idx = getGoodIndex(curplts, funcdist)
        if not idx:
            return

        tmpva = curplts[idx]
        logger.debug("... backwards from... 0x%x", tmpva)
        op = vw.parseOpcode(tmpva)
        stdmnem = op.mnem
        # start there and go backwards
        while tmpva > pltva:
            logger.log(e_cmn.SHITE, "tmpva: 0x%x", tmpva)
            # check if already in a PLT function (ignore if it's part of some other func)
            curfunc = vw.getFunction(tmpva)
            if curfunc is not None and (curfunc == tmpva or isPLT(vw, curfunc)):
                tmpva -= funcdist
                continue

            # check if there's enough room for an even additional one beyond this 
            # (ie. the initial entry in the plt is slightly larger than a normal entry
            leftovers = tmpva - funcdist - pltva
            if leftovers != 0 and leftovers < funcdist:
                logger.debug('skip 0x%x: too close to beginning of PLT', tmpva)
                tmpva -= funcdist 
                continue
            
            # if standard start of plt mnemonic is different...
            op = vw.parseOpcode(tmpva)
            if op.mnem != stdmnem:
                logger.debug('skip 0x%x: WRONG MNEM! (is %r   should be: %r)', tmpva, op.mnem, stdmnem)
                tmpva -= funcdist 
                continue

            logger.info("New PLT Function! 0x%x", tmpva)
            vw.makeFunction(tmpva)

            tmpva -= funcdist 

        # now we go to the end of the PLT!
        tmpva = curplts[idx]
        logger.debug("... forwards from... 0x%x", tmpva)
        endva = pltva + pltsz
        while tmpva < endva:
            logger.log(e_cmn.SHITE, "tmpva: 0x%x", tmpva)
            # check if already in a PLT function
            curfunc = vw.getFunction(tmpva)
            if curfunc is not None and (curfunc == tmpva or isPLT(vw, curfunc)):
                #logger.debug('skip 0x%x: already function', tmpva)
                tmpva += funcdist 
                continue

            # if standard start of plt mnemonic is different...
            op = vw.parseOpcode(tmpva)
            if op.mnem != stdmnem:
                logger.debug('skip 0x%x: WRONG MNEM! (is %r   should be: %r)', tmpva, op.mnem, stdmnem)
                tmpva += funcdist 
                continue

            logger.info("New PLT Function! 0x%x", tmpva)
            vw.makeFunction(tmpva)

            tmpva += funcdist 

def isGOT(vw, va):
    '''
    Check a VA to see if it resides in one of this file's GOT sections
    (uses elfplt.getGOTs()
    '''
    fname = vw.getFileByVa(va)
    gots = elfplt.getGOTs(vw)

    for gotva, gotsz in gots.get(fname):
        if gotva <= va < gotva+gotsz:
            return True

    return False

def isPLT(vw, va):
    '''
    Check a VA to see if it resides in one of this workspace's PLT sections
    (uses elfplt.getPLTs()
    '''
    plts = elfplt.getPLTs(vw)

    for pltva, pltsz in plts:
        if pltva <= va < pltva+pltsz:
            return True

    return False

def getGoodIndex(curplts, funcdist):
    '''
    Roll through a list of PLT entires looking for two that are <funcdist> apart
    '''
    # find starting point of two curplts this distance apart
    idx = None
    for idx in range(1, len(curplts)):
        if curplts[idx] - curplts[idx-1] == funcdist:
            return idx

    return idx

def compareFuncs(vw, fva1, fva2, funcsz):
    '''
    Compare two potential PLT functions... make sure they share a similar pattern.
    They should be nearly identical
    '''
    offset = 0
    while offset < funcsz:
        va1 = fva1 + offset
        va2 = fva2 + offset

        loc1 = vw.getLocation(va1)
        # if loc1 hits a None bail out
        if loc1 is None:
            logger.log(e_cmn.SHITE, '... hit a None location in fva1')
            return False

        lva, lsz, ltype, ltinfo = loc1
        try:
            op1 = vw.parseOpcode(va1)
            op2 = vw.parseOpcode(va2)
       
            if op1.mnem != op2.mnem:
                logger.log(e_cmn.SHITE, "fva1 op mnem (%r) doesn't match fva2 (%r) at offset %d", op1.mnem, op2.mnem, offset)
                return False

            if len(op1.opers) != len(op2.opers):
                logger.log(e_cmn.SHITE, "fva1 op operlen (%r) doesn't match fva2 (%r) at offset %d", op1.opers, op2.opers, offset)
                return False

            for oidx in range(len(op1.opers)):
                oper1 = op1.opers[oidx]
                oper2 = op2.opers[oidx]

                if oper1.__class__ != oper2.__class__:
                    logger.log(e_cmn.SHITE, "fva1 op operclass (%r) doesn't match fva2 (%r) at offset %d", oper1, oper2, offset)
                    return False
                
        except Exception as e:
            # if it fails, we bails
            logger.debug('FAILURE comparing 0x%x and 0x%x: %r', fva1, fva2, e)
            return False

        offset += lsz

    return True




"""
Similar to the memory subsystem, this is a unified way to
access information about objects which contain registers
"""

import envi.exc as e_exc
from envi.const import *

class RegisterContext:

    def __init__(self, regdef=(), metas=(), pcindex=None, spindex=None, srindex=None):
        """
        Hand in a register definition which consists of
        a list of (<name>, <width>) tuples.
        """
        self.loadRegDef(regdef)
        self.loadRegMetas(metas)
        self.setRegisterIndexes(pcindex, spindex, srindex=srindex)

        self._rctx_dirty = False

    def getRegisterSnap(self):
        """
        Use this to bulk save off the register state.
        """
        return list(self._rctx_vals)

    def setRegisterSnap(self, snap):
        """
        Use this to bulk restore the register state.

        NOTE: This may only be used under the assumption that the
              RegisterContext has been initialized the same way
              (like context switches in tracers, or emulaction snaps)
        """
        self._rctx_vals = list(snap)

    def isDirty(self):
        """
        Returns true if registers in this context have been modififed
        since their import.
        """
        return self._rctx_dirty

    def setIsDirty(self, bool):
        self._rctx_dirty = bool

    def setRegisterIndexes(self, pcindex, spindex, srindex=None):
        self._rctx_pcindex = pcindex
        self._rctx_spindex = spindex
        self._rctx_srindex = srindex

    def loadRegDef(self, regdef, defval=0):
        """
        Load a register definition.  A register definition consists
        of a list of tuples with the following format:
        (regname, regwidth)

        NOTE: All widths in envi RegisterContexts are in bits.
        """
        self._rctx_regdef = regdef # Save this for snaps etc..
        self._rctx_names = {}
        self._rctx_ids = {}
        self._rctx_widths = []
        self._rctx_vals  = []
        self._rctx_masks = []

        for i, (name, width) in enumerate(regdef):
            self._rctx_names[name] = i
            self._rctx_ids[i] = name
            self._rctx_widths.append(width)
            self._rctx_masks.append((2**width)-1)
            self._rctx_vals.append(defval)

    def getRegDef(self):
        return self._rctx_regdef

    def loadRegMetas(self, metas, statmetas=None):
        """
        Load a set of defined "meta" registers for this architecture.  Meta
        registers are defined as registers who exist as a subset of the bits
        in some other "real" register. The argument metas is a list of tuples
        with the following format:
        (regname, regidx, reg_shift_offset, reg_width)
        The given example is for the AX register in the i386 subsystem
        regname: "ax"
        reg_shift_offset: 0
        reg_width: 16

        Optionally a set of status meta registers can be loaded as well.
        The argument is a list of tuples with the following format:
        (regname, regidx, reg_shift_offset, reg_width, description)
        """
        self._rctx_regmetas = metas
        for name, idx, offset, width in metas:
            self.addMetaRegister(name, idx, offset, width)

        self._rctx_statmetas = statmetas

    def addMetaRegister(self, name, idx, offset, width):
        """
        Meta registers are registers which are really just directly
        addressable parts of already existing registers (eax -> al).

        To add a meta register, you give the name, the idx of the *real*
        register, the width of the meta reg, and it's left shifted (in bits)
        offset into the real register value.  The RegisterContext will take
        care of accesses after that.
        """
        newidx = (offset << 24) + (width << 16) + idx
        self._rctx_names[name] = newidx
        self._rctx_ids[newidx] = name

    def isMetaRegister(self, index):
        return (index & 0xffff) != index

    def _rctx_Import(self, sobj):
        """
        Given an object with attributes with the same names as
        registers in our context, populate our values from it.

        NOTE: This also clears the dirty flag
        """
        # On import from a structure, we are clean again.
        self._rctx_dirty = False
        for name,idx in self._rctx_names.items():
            # Skip meta registers
            if (idx & 0xffff) != idx:
                continue
            x = getattr(sobj, name, None)
            if x is not None:
                self._rctx_vals[idx] = x

    def _rctx_Export(self, sobj):
        """
        Given an object with attributes with the same names as
        registers in our context, set the ones he has to match
        our values.
        """
        for name,idx in self._rctx_names.items():
            # Skip meta registers
            if (idx & 0xffff) != idx:
                continue
            if hasattr(sobj, name):
                setattr(sobj, name, self._rctx_vals[idx])

    def getRegisterInfo(self, meta=False):
        """
        Return an object which can be stored off, and restored
        to re-initialize a register context.  (much like snapshot
        but it takes the definitions with it)
        """
        regdef = self._rctx_regdef
        regmeta = self._rctx_regmetas
        pcindex = self._rctx_pcindex
        spindex = self._rctx_spindex
        snap = self.getRegisterSnap()

        return (regdef, regmeta, pcindex, spindex, snap)

    def setRegisterInfo(self, info):
        regdef, regmeta, pcindex, spindex, snap = info
        self.loadRegDef(regdef)
        self.loadRegMetas(regmeta)
        self.setRegisterIndexes(pcindex, spindex)
        self.setRegisterSnap(snap)

    def getRegisterName(self, index):
        return self._rctx_ids.get(index,"REG%.8x" % index)

    def getProgramCounter(self):
        """
        Get the value of the program counter for this register context.
        """
        return self.getRegister(self._rctx_pcindex)

    def setProgramCounter(self, value):
        """
        Set the value of the program counter for this register context.
        """
        self.setRegister(self._rctx_pcindex, value)

    def getStackCounter(self):
        return self.getRegister(self._rctx_spindex)

    def setStackCounter(self, value):
        self.setRegister(self._rctx_spindex, value)

    def hasStatusRegister(self):
        '''
        Returns True if this context is aware of a status register.
        '''
        if self._rctx_srindex is None:
            return False

        return True

    def getStatusRegNameDesc(self):
        '''
        Return a list of status register names and descriptions.
        '''
        return [(name, desc) for name, idx, offset, width, desc in self._rctx_statmetas]

    def getStatusRegister(self):
        '''
        Gets the status register for this register context.
        '''
        return self.getRegister(self._rctx_srindex)

    def setStatusRegister(self, value):
        '''
        Sets the status register for this register context.
        '''
        self.setRegister(self._rctx_srindex, value)

    def getStatusFlags(self):
        '''
        Return a dictionary of reg name and reg value for the meta registers
        that are part of the status register.
        '''
        ret = {}
        for name, idx, offset, width, desc in self._rctx_statmetas:
            ret[name] = self.getRegisterByName(name)

        return ret

    def getRegisterByName(self, name):
        idx = self._rctx_names.get(name)
        if idx is None:
            raise e_exc.InvalidRegisterName("Unknown Register: %s" % name)
        return self.getRegister(idx)

    def setRegisterByName(self, name, value):
        idx = self._rctx_names.get(name)
        if idx is None:
            raise e_exc.InvalidRegisterName("Unknown Register: %s" % name)
        self.setRegister(idx, value)

    def getRegisterNames(self):
        '''
        Returns a list of the 'real' (non meta) registers.
        '''
        regs = [rname for rname, ridx in self._rctx_names.items()
                if not self.isMetaRegister(ridx)]
        return regs

    def getRegisterNameIndexes(self):
        '''
        Return a list of all the 'real' (non meta) registers and their indexes.

        Example: for regname, regidx in x.getRegisterNameIndexes():
        '''
        regs = [(rname, ridx) for rname, ridx in self._rctx_names.items()
                if not self.isMetaRegister(ridx)]
        return regs

    def getRegisters(self):
        """
        Get all the *real* registers from this context as a dictionary of name
        value pairs.
        """
        ret = {}
        for name,idx in self._rctx_names.items():
            if (idx & 0xffff) != idx:
                continue
            ret[name] = self.getRegister(idx)
        return ret

    def setRegisters(self, regdict):
        """
        For any name value pairs in the specified dictionary, set the current
        register values in this context.
        """
        for name,value in regdict.items():
            self.setRegisterByName(name, value)

    def getRegisterIndex(self, name):
        """
        Get a register index by name.
        (faster to use the index multiple times)
        """
        return self._rctx_names.get(name)

    def getRegisterWidth(self, index):
        """
        Return the width of the register which lives at the specified
        index (width is always in bits).
        """
        ridx = index & 0xffff
        if ridx == index:
            return self._rctx_widths[index]
        width  = (index >> 16) & 0xff
        return width

    def getRegister(self, index):
        """
        Return the current value of the specified register index.
        """
        ridx = index & 0xffff
        value = self._rctx_vals[ridx]
        if ridx != index:
            value = self._xlateToMetaReg(index, value)
        return value

    def getMetaRegInfo(self, index):
        '''
        Return the appropriate realreg, shift, mask info
        for the specified metareg idx (or None if it's not
        meta).

        Example:
            real_reg, lshift, mask = r.getMetaRegInfo(x)
        '''
        ridx = index & 0xffff
        if ridx == index:
            return None

        offset = (index >> 24) & 0xff
        width  = (index >> 16) & 0xff

        mask = (2**width)-1
        return ridx, offset, mask

    def _xlateToMetaReg(self, index, value):
        '''
        Translate a register value to the meta register value
        (used when getting a meta register)
        '''
        ridx = index & 0xffff
        offset = (index >> 24) & 0xff
        width  = (index >> 16) & 0xff

        mask = (2**width)-1

        if offset != 0:
            value >>= offset

        return value & mask

    def _xlateToNativeReg(self, index, value):
        '''
        Translate a register value to the native register value
        (used when setting a meta register)
        '''
        ridx = index & 0xffff
        width = (index >> 16) & 0xff
        offset = (index >> 24) & 0xff

        # FIXME is it faster to generate or look these up?
        mask = (2 ** width) - 1
        mask = mask << offset

        # NOTE: basewidth is in *bits*
        basewidth = self._rctx_widths[ridx]
        basemask = (2 ** basewidth) - 1

        # cut a whole in basemask at the size/offset of mask
        finalmask = basemask ^ mask

        curval = self._rctx_vals[ridx]

        if offset:
            value <<= offset

        return (value & mask) | (curval & finalmask)

    def setRegister(self, index, value):
        """
        Set a register value by index.
        """
        self._rctx_dirty = True

        ridx = index & 0xffff

        # If it's a meta register index, lets mask it into
        # the real thing...
        if ridx != index:
            value = self._xlateToNativeReg(index, value)

        self._rctx_vals[ridx] = (value & self._rctx_masks[ridx])

    def getRealRegisterNameByIdx(self, regidx):
        """
        Returns the Name of the Containing register (in the case
        of meta-registers) or the name of the register.
        (by Index)
        """
        return self.getRegisterName(regidx & RMETA_NMASK)

    def getRealRegisterName(self, regname):
        """
        Returns the Name of the Containing register (in the case
        of meta-registers) or the name of the register.
        """
        ridx = self.getRegisterIndex(regname)
        if ridx is not None:
            return self.getRegisterName(ridx & RMETA_NMASK)
        return regname


def addLocalEnums(l, regdef):
    """
    Update a dictionary (or module locals) with REG_FOO index
    values for all the base registers defined in regdef.
    """
    for i,(rname,width) in enumerate(regdef):
        if rname is None:   # allow for sparse register banks
            continue

        l["REG_%s" % rname.upper()] = i

def addLocalStatusMetas(l, metas, statmetas, regname):
    '''
    Dynamically create data based on the status register meta register
    definition.
    Adds new meta registers and bitmask constants.
    '''
    for metaname, idx, offset, width, desc in statmetas:
        # create meta registers
        metas.append( (metaname, idx, offset, width) )

        # create local bitmask constants (EFLAGS_%)
        l['%s_%s' % (regname, metaname)] = 1 << offset # TODO: fix for arbitrary width

def addLocalMetas(l, metas):
    """
    Update a dictionary (or module locals) with REG_FOO index
    values for all meta registers defined in metas.
    """
    for name, idx, offset, width in metas:
        l["REG_%s" % name.upper()] = (offset << 24) | (width << 16) | idx



import envi.archs.msp430.emu as e_msp430e
import vivisect.impemu.emulator as v_i_emulator

class Msp430WorkspaceEmulator(v_i_emulator.WorkspaceEmulator, e_msp430e.Msp430Emulator):
    __archemu__ = e_msp430e.Msp430Emulator

    taintregs = [ x for x in range(2, 16) ]

    def __init__(self, vw, **kwargs):
        '''
        Please see the base emulator class in vivisect/impemu/emulator.py for the parameters
        that can be passed through kwargs
        '''
        e_msp430e.Msp430Emulator.__init__(self)
        v_i_emulator.WorkspaceEmulator.__init__(self, vw, **kwargs)

'''
st0len gratuitously from mspgcc:
    Function calling conventions

    Fixed argument lists

    Function arguments are allocated left to right. They are assigned from r15 to r12. If more parameters are passed than will fit in the registers, the rest are passed on the stack. This should be avoided since the code takes a performance hit when using variables residing on the stack.

    Variable argument lists

    Parameters passed to functions that have a variable argument list (printf, scanf, etc.) are all passed on the stack. Any char parameters are extended to ints.

    Return values

    The various functions types return the results as follows:

    char, int and pointer functions return their values r15

    long and float functions return their values in r15:r14

    long long functions return their values r15:r14:r13:r12

    If the returned value wider than 64 bits, it is returned in memory. The first 'hidden' argument to such a function call will be a memory address. All other arguments will be allocated in the usual way, from r14.
    '''



"""
A place for all vivisect contstants so everybody can import them.
"""

# Vivisect workspace events
# NOTE: All events in a vivisect workspace are notified async!
# Also, an event *must* carry with it all the needed information to
# reconstruct the changes to the workspace for the event.
VWE_ADDLOCATION     = 1 # (va,size,ltype,tinfo)
VWE_DELLOCATION     = 2 # (va,size,ltype,tinfo)

VWE_ADDSEGMENT      = 3 # (va, size, name, filename)
VWE_DELSEGMENT      = 4 # FIXME IMPLEMENT

VWE_ADDRELOC        = 5 # (va,rtype)
VWE_DELRELOC        = 6 # # FIXME IMPLEMENT

VWE_ADDMODULE       = 7 # DEPRECATED
VWE_DELMODULE       = 8 # DEPRECATED

VWE_ADDFMODULE      = 9  # DEPRECATED
VWE_DELFMODULE      = 10 # DEPRECATED

VWE_ADDFUNCTION     = 11 # (va, meta)
VWE_DELFUNCTION     = 12 # va

VWE_SETFUNCARGS     = 13 # (fva, arglist)
VWE_SETFUNCMETA     = 14 # (funcva, key, value)

VWE_ADDCODEBLOCK    = 15 # (va, size, funcva)
VWE_DELCODEBLOCK    = 16 # FIXME IMPLEMENT

VWE_ADDXREF         = 17 # (fromva, tova, reftype)
VWE_DELXREF         = 18 # (fromva, tova, reftype)

VWE_SETNAME         = 19 # (va, name)

VWE_ADDMMAP         = 20 # (va, perms, bytes) #OMG MAYBE BIG
VWE_DELMMAP         = 21 # FIXME IMPLEMENT

VWE_ADDEXPORT       = 22 # export object (not for long)
VWE_DELEXPORT       = 23 # export object (not for long)

VWE_SETMETA         = 24 # (key, val)

VWE_COMMENT         = 25 # (va, comment)

VWE_ADDFILE         = 26 # (normname, baseaddr, md5sum)
VWE_DELFILE         = 27  # FIXME IMPLEMENT

VWE_SETFILEMETA     = 28 # (fname, key, value)

VWE_ADDCOLOR        = 29 # (mapname, colordict)
VWE_DELCOLOR        = 30 # mapname

VWE_ADDVASET        = 31 # (name, setdict)
VWE_DELVASET        = 32 # setname

VWE_ADDFREF         = 33 # (va, operidx, value)
VWE_DELFREF         = 34 # FIXME IMPLEMENT

VWE_SETVASETROW     = 35 # (name, rowtup)
VWE_DELVASETROW     = 36 # (name, va)

VWE_ADDFSIG         = 37 # (sigbytes, sigmask)
VWE_DELFSIG         = 38 # FIXME IMPLEMENT

VWE_FOLLOWME        = 39 # LEGACY - not in use.
VWE_CHAT            = 40 # (username, message)

VWE_SYMHINT         = 41 # (va, idx, hint)
VWE_AUTOANALFIN     = 42 # (starttime, endtime)

VWE_WRITEMEM        = 43 # (va, bytes, oldbytes)
VWE_ENDIAN          = 44 # endian

VWE_MAX             = 45

# Constants for vivisect "transient" events which flow through
# the event subsystem but are not recorded to the workspace.
VTE_MASK            = 0x80000000
VTE_IAMLEADER       = 1 # (uuid,user,followname)
VTE_FOLLOWME        = 2 # (uuid,expr)
VTE_KILLLEADER      = 3 # (uuid)
VTE_MODLEADER       = 4 # (uuid,user,followname)
VTE_MAX             = 5

# API fields
API_RET_TYPE    = 0
API_RET_NAME    = 1
API_CCONV       = 2
API_FUNC_NAME   = 3
API_ARG_START   = 4

# Reference Types
# NOTE: All XREFs may have type specific flags
REF_CODE   = 1 # A branch/call
REF_DATA   = 2 # A memory dereference
REF_PTR    = 3 # A pointer immediate (may be in operand *or* part of LOC_PTR)

ref_type_names = {
    REF_CODE: "Code",
    REF_DATA: "Data",
    REF_PTR: "Pointer",
}

#NOTE: The flag values for REF_CODE are the envi.BR_FOO flags
#      which describe opcode branches.

#NOTE: All locations ltypes may not change (backward compat)
LOC_UNDEF   = 0  # An undefined "non-location"
LOC_NUMBER  = 1  # A numerical value (non-pointer)
LOC_STRING  = 2  # A null terminated string
LOC_UNI     = 3  # A null terminated unicode string
LOC_POINTER = 4  # A type to hold a known-derefable pointer that is of appropriate length for arch
LOC_OP      = 5  # An opcode
LOC_STRUCT  = 6  # A custom structure (struct name is in tinfo)
LOC_CLSID   = 7  # A clsid
LOC_VFTABLE = 8  # A c++ vftable
LOC_IMPORT  = 9  # An import dword ptr
LOC_PAD     = 10  # A sequence of bytes which is a pad (string nulls, MS hotpatch... (char is tinfo)
LOC_MAX     = 11

loc_lookups = {v:k for k,v in globals().items() if k.startswith('LOC_')}

loc_type_names = {
        LOC_UNDEF: 'Undefined',
        LOC_NUMBER: 'Num/Int',
        LOC_STRING: 'String',
        LOC_UNI: 'Unicode',
        LOC_POINTER: 'Pointer',
        LOC_OP: 'Opcode',
        LOC_STRUCT: 'Structure',
        LOC_CLSID: 'Clsid',
        LOC_VFTABLE: 'VFTable',
        LOC_IMPORT: 'Import Entry',
        LOC_PAD: 'Pad'
}

# Location tuples contain the following fields indexes.  Many types of
# tuples are used in vivisect for performance.  *most* try to implement
# va and size as the first 2 elements and are then called "Area Compatable"
L_VA    = 0
L_SIZE  = 1
L_LTYPE = 2
L_TINFO = 3

# Code-block tuples are Area Compatable code block descriptions that
# tie code blocks to functions.
CB_VA     = 0
CB_SIZE   = 1
CB_FUNCVA = 2

# Memory Map tuples are Area Compatable tuples that
# describe a loaded memory map
MAP_VA    = 0
MAP_SIZE  = 1
MAP_PERMS = 2
MAP_FNAME = 3

# Segment tuples are Area Compatable tuples that describe
# a "section" or "segment" inside a memory map.
SEG_VA    = 0
SEG_SIZE  = 1
SEG_NAME  = 2 # The name of the segment ".text" ".plt"
SEG_FNAME = 3 # The *normalized* name of the file

# XREF tuples *not* area compatable tuples
XR_FROM  = 0
XR_TO    = 1
XR_RTYPE = 2
XR_RFLAG = 3

# Export Types
EXP_UNTYPED  = 0xffffffff
EXP_FUNCTION = 0
EXP_DATA     = 1

# Relocation types
RTYPE_BASERELOC = 0 # VA contains a pointer to a va (and is assumed fixed up by parser)
RTYPE_BASEOFF   = 1 # Add Base and Offset to a pointer at a memory location
RTYPE_BASEPTR   = 2 # Like BASEOFF, but treated as a Pointer, not part of an instruction/etc.

REBASE_TYPES = (RTYPE_BASEOFF, RTYPE_BASEPTR)

# Function Local Symbol Types
LSYM_NAME   = 0 # syminfo is a (typestr,name) tuple
LSYM_FARG   = 1 # syminfo is an argument index


# vaset "type" constants
VASET_ADDRESS   = 0
VASET_INTEGER   = 1
VASET_STRING    = 2
VASET_HEXTUP    = 3
VASET_COMPLEX   = 4

# Symboliks effect types
EFFTYPE_DEBUG        = 0
EFFTYPE_SETVAR       = 1
EFFTYPE_READMEM      = 2
EFFTYPE_WRITEMEM     = 3
EFFTYPE_CALLFUNC     = 4
EFFTYPE_CONSTRAIN    = 5

# symboliks object types
SYMT_VAR            = 0
SYMT_ARG            = 1
SYMT_CALL           = 2
SYMT_MEM            = 3
SYMT_SEXT           = 4
SYMT_CONST          = 5
SYMT_LOOKUP         = 6
SYMT_NOT            = 7

SYMT_OPER           = 0x00010000
SYMT_OPER_ADD       = SYMT_OPER | 1
SYMT_OPER_SUB       = SYMT_OPER | 2
SYMT_OPER_MUL       = SYMT_OPER | 3
SYMT_OPER_DIV       = SYMT_OPER | 4
SYMT_OPER_AND       = SYMT_OPER | 5
SYMT_OPER_OR        = SYMT_OPER | 6
SYMT_OPER_XOR       = SYMT_OPER | 7
SYMT_OPER_MOD       = SYMT_OPER | 8
SYMT_OPER_LSHIFT    = SYMT_OPER | 9
SYMT_OPER_RSHIFT    = SYMT_OPER | 10
SYMT_OPER_POW       = SYMT_OPER | 11

SYMT_CON            = 0x00020000
SYMT_CON_EQ         = SYMT_CON | 1
SYMT_CON_NE         = SYMT_CON | 2
SYMT_CON_GT         = SYMT_CON | 3
SYMT_CON_GE         = SYMT_CON | 4
SYMT_CON_LT         = SYMT_CON | 5
SYMT_CON_LE         = SYMT_CON | 6
SYMT_CON_UNK        = SYMT_CON | 7
SYMT_CON_NOTUNK     = SYMT_CON | 8



'''
A module with some cute toys for monitoring allocations.
'''

import vtrace

import envi.archs.i386 as e_i386


class ReturnBreak(vtrace.Breakpoint):
    def __init__(self, addr, chsize, chflags):
        vtrace.Breakpoint.__init__(self, addr)
        self.fastbreak = True
        self._chsize = chsize
        self._chflags = chflags

    def notify(self, event, trace):
        eax = trace.getRegister(e_i386.REG_EAX)
        a = trace.getMeta('HeapAllocs')
        a.append((self.address, eax, self._chsize, self._chflags))
        trace.runAgain()


class RtlAllocateHeapBreak(vtrace.Breakpoint):

    def __init__(self, addr):
        vtrace.Breakpoint.__init__(self, addr)
        self.fastbreak = True

    def notify(self, event, trace):

        sp = trace.getStackCounter()
        (saved_eip,
         heap,
         flags,
         size) = trace.readMemoryFormat(sp, '<4P')

        if trace.getBreakpointByAddr(saved_eip) is None:
            bp = ReturnBreak(saved_eip, size, flags)
            trace.addBreakpoint(bp)

        trace.runAgain()


def watchHeapAllocs(trace):
    '''
    Add a breakpoint to ntdll.RtlAllocateHeap to watch for
    allocations and track who made them...
    '''
    clearHeapAllocs(trace)
    addr = trace.parseExpression('ntdll.RtlAllocateHeap')
    bp = RtlAllocateHeapBreak(addr)
    trace.addBreakpoint(bp)


def clearHeapAllocs(trace):
    trace.setMeta('HeapAllocs', [])


def getHeapAllocs(trace):
    '''
    Return a list of (caller_eip, heap_chunk, size, flags) tuples
    '''
    return trace.getMeta('HeapAllocs', [])



import visgraph.renderers as vg_render

example = '''
<?xml version="1.0" standalone="no"?>
<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 1.1//EN"
"http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd">

<svg width="100%" height="100%" version="1.1"
xmlns="http://www.w3.org/2000/svg">

<rect x="20" y="20" rx="20" ry="20" width="250" height="100"
style="fill:red;stroke:black;stroke-width:5;opacity:0.5"/>
</svg>
'''

class SvgGraphRenderer(vg_render.GraphRenderer):

    def __init__(self, graph, svgfile):
        vg_render.GraphRenderer.__init__(self, graph)
        self.svgfile = svgfile
        self._node_xml = []
        self._edge_xml = []

    def renderNode(self, nid, ninfo, xpos, ypos):
        rectstr = '<rect x="%d" y="%d" width="10" height="10" style="fill:red;stroke:black;stroke-width:5;opacity:0.5"/>' % (xpos,ypos)
        self._node_xml.append(rectstr)

    def renderEdge(self, eid, einfo, points):
        pointstr = ' '.join([ '%d,%d' % (x,y) for x,y in points ])
        edgestr = '<polyline points="%s" style="fill:white;stroke:red;stroke-width:2"/>' % pointstr
        self._edge_xml.append(edgestr)

    def setNodeSizes(self, graph):
        # Do the node sizing...
        for nid,ninfo in graph.getNodes():
            ninfo['size'] = (10,10)

    def endRender(self):
        # Actually write out the file.
        with open(self.svgfile, 'wb') as f:
            f.write('<?xml version="1.0" standalone="no"?>\n')
            #f.write('<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 1.1//EN http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd" >\n')
            xsize = self._vg_canvas_width
            ysize = self._vg_canvas_height
            f.write('<svg width="%d" height="%d" version="1.1" xmlns="http://www.w3.org/2000/svg">\n' % (xsize,ysize))
            for x in self._node_xml:
                f.write(x + '\n')
            for x in self._edge_xml:
                f.write(x + '\n')
            f.write('</svg>\n')

if __name__ == '__main__':

    import visgraph.graphcore as vg_graphcore
    import visgraph.layouts.force as vg_force
    import visgraph.renderers.svgrend as vg_svgrend

    g = vg_graphcore.HierGraph()

    g.addNode('A', rootnode=True)
    g.addNode('B')
    g.addNode('C')
    g.addNode('D')

    g.addNode('E')
    g.addNode('F')
    g.addNode('G')

    g.addEdgeByNids('A','B')
    g.addEdgeByNids('A','C')

    g.addEdgeByNids('B','D')
    g.addEdgeByNids('B','E')

    g.addEdgeByNids('C','F')
    g.addEdgeByNids('C','G')

    #layout = vg_dynadag.DynadagLayout(g)
    layout = vg_force.ForceLayout(g)
    rend = vg_svgrend.SvgGraphRenderer(g, 'test.svg')

    layout.renderGraph(rend)



"""
Platform Support Modules

The vtrace platforms package is where the modules live that
implement platform specific functionality that is used behind
the scenes of the top level Trace object.  If you're not porting
vtrace to a new platform, or chasing a bug, you probably don't need
anything from here...
"""
# Copyright (C) 2007 Invisigoth - See LICENSE file for details



import logging
import unittest

import PE
import envi.const as e_const

import vivisect
import vivisect.const as viv_con
import vivisect.parsers.pe

import vivisect.tests.helpers as helpers
import vivisect.tests.utils as v_t_utils

logger = logging.getLogger(__name__)


class PETests(v_t_utils.VivTest):

    @classmethod
    def setUpClass(cls):
        super(PETests, cls).setUpClass()
        cls.psexec_fn = helpers.getTestPath('windows', 'i386', 'PsExec.exe')

        cls.vw_psexec = helpers.getTestWorkspace('windows', 'i386', 'PsExec.exe')
        cls.vw_sphinx = helpers.getTestWorkspace('windows', 'i386', 'sphinx_livepretend.exe')

        cls.vw_mimi = vivisect.VivWorkspace()
        mimi_fn = helpers.getTestPath('windows', 'i386', 'mimikatz.exe_')
        cls.vw_mimi.loadFromFile(mimi_fn)

        # this binary is a little big (1MB)
        # and we only care about the delay import table
        # so, don't do a full analysis
        fn_471 = helpers.getTestPath('windows', 'i386', '471ce36855fec6b44398b9b1e3cfb9e74b122fb2cc20fdf6603ebda39f86dddf')
        cls.vw_471 = vivisect.VivWorkspace()
        vivisect.parsers.pe.parseFile(cls.vw_471, fn_471)

    def test_function_disasm(self):
        disasm = [
            (0x4028e0, 1, 'push ebp'),
            (0x4028e1, 2, 'mov ebp,esp'),
            (0x4028e3, 3, 'sub esp,8'),
            (0x4028e6, 1, 'push esi'),
            (0x4028e7, 3, 'mov esi,dword [ebp + 8]'),
            (0x4028ea, 1, 'push edi'),
            (0x4028eb, 2, 'push 64'),
            (0x4028ed, 1, 'push esi'),
            (0x4028ee, 6, 'call dword [0x0041a168]'),
            (0x4028f4, 5, 'push 0x00420944'),
            (0x4028f9, 5, 'push 0x0042095c'),
            (0x4028fe, 6, 'call dword [0x0041a178]'),
            (0x402904, 1, 'push eax'),
            (0x402905, 6, 'call dword [0x0041a18c]'),
            (0x40290b, 2, 'mov edi,eax'),
            (0x40290d, 2, 'test edi,edi'),
            (0x40290f, 2, 'jz 0x00402935'),
        ]

        cbva = 0x4028e0
        self.assertEqual(self.vw_psexec.getFunction(cbva), cbva)
        for (va, size, inst) in disasm:
            lva, lsize, ltype, linfo = self.vw_psexec.getLocation(cbva)
            op = self.vw_psexec.parseOpcode(lva)
            self.assertEqual(ltype, viv_con.LOC_OP)
            self.assertEqual(size, lsize)
            self.assertEqual(str(op), inst)
            cbva += size

    def test_export_by_name(self):
        file_path = helpers.getTestPath('windows', 'i386', 'export_by_name.dll')
        pe = PE.peFromFileName(file_path)
        export_list = pe.getExports()
        self.assertEqual(len(export_list), 2, "expecting 2 exported functions")
        self.assertEqual(export_list[0][1], 0, "exported function with ordinal 0 not found")
        self.assertEqual(export_list[0][2], "Func1", "exported function with name 'Func1' not found")
        self.assertEqual(export_list[1][1], 1, "exported function with ordinal 1 not found")
        self.assertEqual(export_list[1][2], "Func2", "exported function with name 'Func2' not found")

    def test_export_by_ordinal_base_01(self):
        file_path = helpers.getTestPath('windows', 'i386', 'export_by_ordinal_base_01.dll')
        pe = PE.peFromFileName(file_path)
        export_list = pe.getExports()
        self.assertEqual(len(export_list), 2, "expecting 2 exported functions")
        self.assertEqual(export_list[0][1], 1, "exported function with ordinal 1 not found")
        self.assertEqual(export_list[1][1], 2, "exported function with ordinal 2 not found")

    def test_export_by_ordinal_base_45(self):
        file_path = helpers.getTestPath('windows', 'i386', 'export_by_ordinal_base_45.dll')
        pe = PE.peFromFileName(file_path)
        export_list = pe.getExports()
        self.assertEqual(len(export_list), 2, "expecting 2 exported functions")
        self.assertEqual(export_list[0][1], 45, "exported function with ordinal 45 not found")
        self.assertEqual(export_list[1][1], 55, "exported function with ordinal 55 not found")

    def test_pe_metainfo(self):
        self.assertEqual(self.vw_psexec.getMeta('Architecture'), 'i386')
        self.assertEqual(self.vw_psexec.getMeta('DefaultCall'), 'cdecl')
        self.assertEqual(self.vw_psexec.getMeta('Format'), 'pe')
        self.assertEqual(self.vw_psexec.getMeta('Platform'), 'windows')
        self.assertEqual(self.vw_psexec.getMeta('StorageModule'), 'vivisect.storage.basicfile')
        self.assertEqual(self.vw_psexec.getMeta('StorageName'), self.psexec_fn + '.viv')

        self.assertEqual(len(self.vw_psexec.getMeta('NoReturnApis')), 5)
        self.assertTrue(self.vw_psexec.metadata['NoReturnApis']['kernel32.exitprocess'])
        self.assertTrue(self.vw_psexec.metadata['NoReturnApis']['kernel32.exitthread'])
        self.assertTrue(self.vw_psexec.metadata['NoReturnApis']['kernel32.fatalexit'])
        self.assertTrue(self.vw_psexec.metadata['NoReturnApis']['ntdll.rtlexituserthread'])
        self.assertTrue(self.vw_psexec.metadata['NoReturnApis']['ntoskrnl.kebugcheckex'])

        noret = [
            0x41a1c0,  # ExitProcess
            0x412380,  # calls 0x407027
            0x409143,  # legit no return
            0x40ba2b,  # ends in int3 only
            0x407027,  # calls 0x4070e5 and then in3
            0x412389,  # every path ends in int3
            0x40ba07,  # ends in call to 0x412389 and int3
            0x4018d0,  # ends in call to 0x4072ca and int3
            0x407011,  # Ends in a call to ExitProcess
            0x409be3,  # Calls Exit thread, but also ends in a call to 0x409b77 followed by cccccccc bytes
            0x41a1f4,  # Actual ExitThread import
            0x409bb8,  # Ends in a call to ExitThread and then int3
            0x4090bd,  # Ends in a call to ExitThread and then int3
            0x409dbf,  # Ends in a call to 0x407011
        ]
        meta = self.vw_psexec.getMeta('NoReturnApisVa')
        for nva in noret:
            self.assertTrue(nva in meta, msg='0x%.8x is not a no return va!' % nva)
        self.assertTrue(self.vw_psexec.metadata['NoReturnApisVa'][4301248])
        self.assertTrue(self.vw_psexec.metadata['NoReturnApisVa'][4301300])

    def test_pe_filemeta(self):
        vw = self.vw_psexec
        files = vw.getFiles()
        self.assertTrue('psexec' in files)
        self.assertEqual('27304b246c7d5b4e149124d5f93c5b01', vw.getFileMeta('psexec', 'md5sum'))
        self.assertEqual('3337E3875B05E0BFBA69AB926532E3F179E8CFBF162EBB60CE58A0281437A7EF',
                         vw.getFileMeta('psexec', 'sha256'))
        self.assertEqual('2.2', vw.getFileMeta('psexec', 'Version'))
        self.assertEqual(0x400000, vw.getFileMeta('psexec', 'imagebase'))
        self.assertTrue(vw.getFileMeta('psexec', 'SafeSEH'))

    def test_delay_imports(self):
        vw = self.vw_psexec
        delays = sorted(vw.getVaSetRows('DelayImports'))
        ans = [
            (0x42b2ec, 'gdi32.EndPage'),
            (0x42b2f0, 'gdi32.EndDoc'),
            (0x42b2f4, 'gdi32.StartDocW'),
            (0x42b2f8, 'gdi32.SetMapMode'),
            (0x42b2fc, 'gdi32.GetDeviceCaps'),
            (0x42b300, 'gdi32.StartPage'),
            (0x42b308, 'user32.GetSysColorBrush'),
            (0x42b30c, 'user32.SetCursor'),
            (0x42b310, 'user32.LoadCursorW'),
            (0x42b314, 'user32.LoadStringW'),
            (0x42b318, 'user32.SetWindowTextW'),
            (0x42b31c, 'user32.GetDlgItem'),
            (0x42b320, 'user32.EndDialog'),
            (0x42b324, 'user32.DialogBoxIndirectParamW'),
            (0x42b328, 'user32.InflateRect'),
            (0x42b32c, 'user32.SendMessageW'),
        ]
        self.assertEqual(delays, ans)
 
    def test_delay_imports_with_va(self):
        vw = self.vw_471
        delays = sorted(vw.getVaSetRows('DelayImports'))
        ans = [
            (0x4e3a34, 'advapi32.RegCloseKey'),
            (0x4e3a3c, 'gdi32.DeleteObject'),
            (0x4e3a40, 'gdi32.CreateRectRgn'),
            (0x4e3a44, 'gdi32.GetPixel'),
            (0x4e3a48, 'gdi32.SelectObject'),
            (0x4e3a4c, 'gdi32.CreateCompatibleDC'),
            (0x4e3a50, 'gdi32.DeleteDC'),
            (0x4e3a54, 'gdi32.CombineRgn'),
            (0x4e3a58, 'gdi32.OffsetRgn'),
            (0x4e3a5c, 'gdi32.LPtoDP'),
            (0x4e3a60, 'gdi32.DPtoLP'),
            (0x4e3a64, 'gdi32.CreateEllipticRgn'),
            (0x4e3a68, 'gdi32.CreateRectRgnIndirect'),
            (0x4e3a6c, 'gdi32.GetRgnBox'),
            (0x4e3a70, 'gdi32.SelectPalette'),
            (0x4e3a74, 'gdi32.SetMapMode'),
            (0x4e3a78, 'gdi32.RealizePalette'),
            (0x4e3a7c, 'gdi32.GetDIBits'),
            (0x4e3a80, 'gdi32.GetDeviceCaps'),
            (0x4e3a88, 'redeye.ARER_AutoDetect'),
            (0x4e3a8c, 'redeye.ARER_Correct'),
            (0x4e3a94, 'winmm.timeGetTime'),
            (0x4e3a98, 'winmm.waveOutGetNumDevs'),
            (0x4e3aa0, 'magfileio.magIOGetSaveList'),
            (0x4e3aa4, 'magfileio.magIOLoadPlugsEx'),
            (0x4e3aa8, 'magfileio.magIOUnloadPlugs'),
            (0x4e3ab0, 'spotremove.pmi_spot_remove_auto'),
            (0x4e3ab4, 'spotremove.pmi_red_eye_remove_auto'),
        ]
        self.assertEqual(delays, ans)
 
    def test_imports(self):
        vw = self.vw_psexec
        imps = vw.getImports()
        ans = [(4301428, 4, 9, 'version.GetFileVersionInfoSizeW'), (4301432, 4, 9, 'version.GetFileVersionInfoW'),
               (4301436, 4, 9, 'version.VerQueryValueW'), (4301416, 4, 9, 'netapi32.NetServerEnum'),
               (4301420, 4, 9, 'netapi32.NetApiBufferFree'), (4301444, 4, 9, 'ws2_32.gethostname'),
               (4301448, 4, 9, 'ws2_32.WSAStartup'), (4301452, 4, 9, 'ws2_32.inet_ntoa'),
               (4301456, 4, 9, 'ws2_32.gethostbyname'), (4301404, 4, 9, 'mpr.WNetCancelConnection2W'),
               (4301408, 4, 9, 'mpr.WNetAddConnection2W'), (4300984, 4, 9, 'kernel32.GetExitCodeProcess'),
               (4300988, 4, 9, 'kernel32.ResumeThread'), (4300992, 4, 9, 'kernel32.WaitForMultipleObjects'),
               (4300996, 4, 9, 'kernel32.GetFileTime'), (4301000, 4, 9, 'kernel32.DuplicateHandle'),
               (4301004, 4, 9, 'kernel32.DisconnectNamedPipe'), (4301008, 4, 9, 'kernel32.SetNamedPipeHandleState'),
               (4301012, 4, 9, 'kernel32.TransactNamedPipe'), (4301016, 4, 9, 'kernel32.CreateEventW'),
               (4301020, 4, 9, 'kernel32.GetCurrentProcessId'), (4301024, 4, 9, 'kernel32.GetFullPathNameW'),
               (4301028, 4, 9, 'kernel32.SetFileAttributesW'), (4301032, 4, 9, 'kernel32.GetFileAttributesW'),
               (4301036, 4, 9, 'kernel32.CopyFileW'), (4301040, 4, 9, 'kernel32.WaitNamedPipeW'),
               (4301044, 4, 9, 'kernel32.SetConsoleCtrlHandler'), (4301048, 4, 9, 'kernel32.SetConsoleTitleW'),
               (4301052, 4, 9, 'kernel32.ReadConsoleW'), (4301056, 4, 9, 'kernel32.GetVersion'),
               (4301060, 4, 9, 'kernel32.SetProcessAffinityMask'), (4301064, 4, 9, 'kernel32.ReadFile'),
               (4301068, 4, 9, 'kernel32.GetConsoleScreenBufferInfo'), (4301072, 4, 9, 'kernel32.MultiByteToWideChar'),
               (4301076, 4, 9, 'kernel32.GetComputerNameW'), (4301080, 4, 9, 'kernel32.DeleteFileW'),
               (4301084, 4, 9, 'kernel32.CreateFileW'), (4301088, 4, 9, 'kernel32.GetSystemDirectoryW'),
               (4301092, 4, 9, 'kernel32.FindResourceW'), (4301096, 4, 9, 'kernel32.LoadLibraryExW'),
               (4301100, 4, 9, 'kernel32.FormatMessageA'), (4301104, 4, 9, 'kernel32.GetTickCount'),
               (4301108, 4, 9, 'kernel32.CloseHandle'), (4301112, 4, 9, 'kernel32.WriteFile'),
               (4301116, 4, 9, 'kernel32.SizeofResource'), (4301120, 4, 9, 'kernel32.LoadResource'),
               (4301124, 4, 9, 'kernel32.Sleep'), (4301128, 4, 9, 'kernel32.WaitForSingleObject'),
               (4301132, 4, 9, 'kernel32.SetEndOfFile'), (4301136, 4, 9, 'kernel32.SetEvent'),
               (4301140, 4, 9, 'kernel32.SetLastError'), (4301144, 4, 9, 'kernel32.GetLastError'),
               (4301148, 4, 9, 'kernel32.GetCurrentProcess'), (4301152, 4, 9, 'kernel32.FreeLibrary'),
               (4301156, 4, 9, 'kernel32.LockResource'), (4301160, 4, 9, 'kernel32.SetPriorityClass'),
               (4301164, 4, 9, 'kernel32.GetModuleFileNameW'), (4301168, 4, 9, 'kernel32.GetCommandLineW'),
               (4301172, 4, 9, 'kernel32.GetModuleHandleW'), (4301176, 4, 9, 'kernel32.LoadLibraryW'),
               (4301180, 4, 9, 'kernel32.GetStdHandle'), (4301184, 4, 9, 'kernel32.GetFileType'),
               (4301188, 4, 9, 'kernel32.LocalFree'), (4301192, 4, 9, 'kernel32.LocalAlloc'),
               (4301196, 4, 9, 'kernel32.GetProcAddress'), (4301200, 4, 9, 'kernel32.FreeEnvironmentStringsW'),
               (4301204, 4, 9, 'kernel32.LCMapStringW'), (4301208, 4, 9, 'kernel32.OutputDebugStringW'),
               (4301212, 4, 9, 'kernel32.HeapSize'), (4301216, 4, 9, 'kernel32.HeapReAlloc'),
               (4301220, 4, 9, 'kernel32.SetFilePointerEx'), (4301224, 4, 9, 'kernel32.WriteConsoleW'),
               (4301228, 4, 9, 'kernel32.GetEnvironmentVariableW'), (4301232, 4, 9, 'kernel32.RaiseException'),
               (4301236, 4, 9, 'kernel32.LoadLibraryExA'), (4301240, 4, 9, 'kernel32.EncodePointer'),
               (4301244, 4, 9, 'kernel32.DecodePointer'), (4301248, 4, 9, 'kernel32.ExitProcess'),
               (4301252, 4, 9, 'kernel32.GetModuleHandleExW'), (4301256, 4, 9, 'kernel32.WideCharToMultiByte'),
               (4301260, 4, 9, 'kernel32.HeapFree'), (4301264, 4, 9, 'kernel32.HeapAlloc'),
               (4301268, 4, 9, 'kernel32.GetConsoleMode'), (4301272, 4, 9, 'kernel32.ReadConsoleInputA'),
               (4301276, 4, 9, 'kernel32.SetConsoleMode'), (4301280, 4, 9, 'kernel32.EnterCriticalSection'),
               (4301284, 4, 9, 'kernel32.LeaveCriticalSection'), (4301288, 4, 9, 'kernel32.SetStdHandle'),
               (4301292, 4, 9, 'kernel32.CreateThread'), (4301296, 4, 9, 'kernel32.GetCurrentThreadId'),
               (4301300, 4, 9, 'kernel32.ExitThread'), (4301304, 4, 9, 'kernel32.IsDebuggerPresent'),
               (4301308, 4, 9, 'kernel32.IsProcessorFeaturePresent'), (4301312, 4, 9, 'kernel32.GetStringTypeW'),
               (4301316, 4, 9, 'kernel32.IsValidCodePage'), (4301320, 4, 9, 'kernel32.GetACP'),
               (4301324, 4, 9, 'kernel32.GetOEMCP'), (4301328, 4, 9, 'kernel32.GetCPInfo'),
               (4301332, 4, 9, 'kernel32.DeleteCriticalSection'), (4301336, 4, 9, 'kernel32.UnhandledExceptionFilter'),
               (4301340, 4, 9, 'kernel32.SetUnhandledExceptionFilter'), (4301344, 4, 9, 'kernel32.InitializeCriticalSectionAndSpinCount'),
               (4301348, 4, 9, 'kernel32.TerminateProcess'), (4301352, 4, 9, 'kernel32.TlsAlloc'),
               (4301356, 4, 9, 'kernel32.TlsGetValue'), (4301360, 4, 9, 'kernel32.TlsSetValue'),
               (4301364, 4, 9, 'kernel32.TlsFree'), (4301368, 4, 9, 'kernel32.GetStartupInfoW'),
               (4301372, 4, 9, 'kernel32.GetProcessHeap'), (4301376, 4, 9, 'kernel32.FlushFileBuffers'),
               (4301380, 4, 9, 'kernel32.GetConsoleCP'), (4301384, 4, 9, 'kernel32.RtlUnwind'),
               (4301388, 4, 9, 'kernel32.QueryPerformanceCounter'), (4301392, 4, 9, 'kernel32.GetSystemTimeAsFileTime'),
               (4301396, 4, 9, 'kernel32.GetEnvironmentStringsW'), (4300976, 4, 9, 'comdlg32.PrintDlgW'),
               (4300800, 4, 9, 'advapi32.LsaClose'), (4300804, 4, 9, 'advapi32.CreateProcessAsUserW'),
               (4300808, 4, 9, 'advapi32.CryptHashData'), (4300812, 4, 9, 'advapi32.CryptCreateHash'),
               (4300816, 4, 9, 'advapi32.CryptDecrypt'), (4300820, 4, 9, 'advapi32.CryptEncrypt'),
               (4300824, 4, 9, 'advapi32.CryptImportKey'), (4300828, 4, 9, 'advapi32.CryptExportKey'),
               (4300832, 4, 9, 'advapi32.CryptDestroyKey'), (4300836, 4, 9, 'advapi32.CryptDeriveKey'),
               (4300840, 4, 9, 'advapi32.CryptGenKey'), (4300844, 4, 9, 'advapi32.CryptReleaseContext'),
               (4300848, 4, 9, 'advapi32.CryptAcquireContextW'), (4300852, 4, 9, 'advapi32.StartServiceW'),
               (4300856, 4, 9, 'advapi32.QueryServiceStatus'), (4300860, 4, 9, 'advapi32.OpenServiceW'),
               (4300864, 4, 9, 'advapi32.OpenSCManagerW'), (4300868, 4, 9, 'advapi32.DeleteService'),
               (4300872, 4, 9, 'advapi32.CreateServiceW'), (4300876, 4, 9, 'advapi32.ControlService'),
               (4300880, 4, 9, 'advapi32.CloseServiceHandle'), (4300884, 4, 9, 'advapi32.OpenProcessToken'),
               (4300888, 4, 9, 'advapi32.LsaEnumerateAccountRights'), (4300892, 4, 9, 'advapi32.LsaOpenPolicy'),
               (4300896, 4, 9, 'advapi32.LsaFreeMemory'), (4300900, 4, 9, 'advapi32.SetSecurityInfo'),
               (4300904, 4, 9, 'advapi32.GetSecurityInfo'), (4300908, 4, 9, 'advapi32.LookupPrivilegeValueW'),
               (4300912, 4, 9, 'advapi32.AddAccessAllowedAce'), (4300916, 4, 9, 'advapi32.GetAce'),
               (4300920, 4, 9, 'advapi32.AddAce'), (4300924, 4, 9, 'advapi32.InitializeAcl'),
               (4300928, 4, 9, 'advapi32.GetLengthSid'), (4300932, 4, 9, 'advapi32.FreeSid'),
               (4300936, 4, 9, 'advapi32.AllocateAndInitializeSid'), (4300940, 4, 9, 'advapi32.SetTokenInformation'),
               (4300944, 4, 9, 'advapi32.GetTokenInformation'), (4300948, 4, 9, 'advapi32.RegSetValueExW'),
               (4300952, 4, 9, 'advapi32.RegQueryValueExW'), (4300956, 4, 9, 'advapi32.RegOpenKeyExW'),
               (4300960, 4, 9, 'advapi32.RegOpenKeyW'), (4300964, 4, 9, 'advapi32.RegCreateKeyW'),
               (4300968, 4, 9, 'advapi32.RegCloseKey'), (4371208, 4, 9, 'user32.GetSysColorBrush'),
               (4371212, 4, 9, 'user32.SetCursor'), (4371216, 4, 9, 'user32.LoadCursorW'),
               (4371220, 4, 9, 'user32.LoadStringW'), (4371224, 4, 9, 'user32.SetWindowTextW'),
               (4371228, 4, 9, 'user32.GetDlgItem'), (4371232, 4, 9, 'user32.EndDialog'),
               (4371236, 4, 9, 'user32.DialogBoxIndirectParamW'), (4371240, 4, 9, 'user32.InflateRect'),
               (4371244, 4, 9, 'user32.SendMessageW'), (4371180, 4, 9, 'gdi32.EndPage'),
               (4371184, 4, 9, 'gdi32.EndDoc'), (4371188, 4, 9, 'gdi32.StartDocW'),
               (4371192, 4, 9, 'gdi32.SetMapMode'), (4371196, 4, 9, 'gdi32.GetDeviceCaps'),
               (4371200, 4, 9, 'gdi32.StartPage')]

        self.assertEqual(sorted(imps), sorted(ans))

    def test_pe_jmptable(self):
        vw = self.vw_sphinx

        tblins = 0x00403b3c
        xrefs = vw.getXrefsFrom(tblins, viv_con.REF_PTR)
        self.assertEqual(len(xrefs), 1)
        self.assertEqual(xrefs[0], (tblins, 0x403bc4, 3, 0))

        xrefs = vw.getXrefsFrom(tblins, viv_con.REF_CODE)
        self.assertEqual(len(xrefs), 4)
        codeblocks = [
            (4209468, 4209475, 1, 2),
            (4209468, 4209509, 1, 2),
            (4209468, 4209517, 1, 2),
            (4209468, 4209525, 1, 2)
        ]
        for xr in xrefs:
            self.assertTrue(xr in codeblocks)

    def test_pe_dynamic_noret(self):
        vw = self.vw_sphinx
        noret = [0x408d40, 0x408da0, 0x429229, 0x426e8a, 0x43b06c, 0x4268d4, 0x4268ba, 0x4291fd]
        for va in noret:
            self.assertTrue(vw.isNoReturnVa(va), msg='0x%.8x is not no return!' % va)

    def test_emulation_vaset(self):
        vw = self.vw_psexec
        funcs = set([4281345, 4294188, 4289086, 4293695, 4282944, 4290121, 4198480, 4290145, 4210288, 4282482, 4280472, 4268701, 4208800, 4210848, 4290211, 4209328, 4291252, 4290228, 4289721, 4239035, 4293319, 4281548, 4287692, 4288719, 4293840, 4221654, 4233944, 4221660, 4207328, 4209376, 4221666, 4291296, 4221672, 4280554, 4288749, 4287725, 4281584, 4259060, 4289292, 4282646, 4285732, 4216112, 4291392, 4280642, 4290370, 4206416, 4259158, 4208992, 4209504, 4280685, 4290929, 4287859, 4293495, 4208512, 4259207, 4211088, 4293008, 4199824, 4280730, 4241818, 4259229, 4211616, 4282788, 4294064, 4284338, 4289976, 4209088, 4209600, 4289993, 4198864, 4213200, 4206032, 4259282, 4209120, 4209632, 4231139, 4269557, 4289528])
        for fva in funcs:
            self.assertEqual(fva, vw.getFunction(fva))

        # if we reorder analysis passes or new ones, this might change, and that's okay.
        # this is more designed to be a smoke test that we actually get a populated vaset
        emufuncs = vw.getVaSet('EmucodeFunctions')
        self.assertTrue(len(emufuncs) > 0)
        e = set(emufuncs.keys())
        f = set(funcs)
        self.assertTrue(f.intersection(e) == f)

    def test_get_imports(self):
        vw = self.vw_psexec
        imps = set([(4301428, 4, 9, 'version.GetFileVersionInfoSizeW'),
                    (4301432, 4, 9, 'version.GetFileVersionInfoW'),
                    (4301436, 4, 9, 'version.VerQueryValueW'),
                    (4301416, 4, 9, 'netapi32.NetServerEnum'),
                    (4301420, 4, 9, 'netapi32.NetApiBufferFree'),
                    (4301444, 4, 9, 'ws2_32.gethostname'),
                    (4301448, 4, 9, 'ws2_32.WSAStartup'),
                    (4301452, 4, 9, 'ws2_32.inet_ntoa'),
                    (4301456, 4, 9, 'ws2_32.gethostbyname'),
                    (4301404, 4, 9, 'mpr.WNetCancelConnection2W'),
                    (4301408, 4, 9, 'mpr.WNetAddConnection2W'),
                    (4300984, 4, 9, 'kernel32.GetExitCodeProcess'),
                    (4300988, 4, 9, 'kernel32.ResumeThread'),
                    (4300992, 4, 9, 'kernel32.WaitForMultipleObjects'),
                    (4300996, 4, 9, 'kernel32.GetFileTime'),
                    (4301000, 4, 9, 'kernel32.DuplicateHandle'),
                    (4301004, 4, 9, 'kernel32.DisconnectNamedPipe'),
                    (4301008, 4, 9, 'kernel32.SetNamedPipeHandleState'),
                    (4301012, 4, 9, 'kernel32.TransactNamedPipe'),
                    (4301016, 4, 9, 'kernel32.CreateEventW'),
                    (4301020, 4, 9, 'kernel32.GetCurrentProcessId'),
                    (4301024, 4, 9, 'kernel32.GetFullPathNameW'),
                    (4301028, 4, 9, 'kernel32.SetFileAttributesW'),
                    (4301032, 4, 9, 'kernel32.GetFileAttributesW'),
                    (4301036, 4, 9, 'kernel32.CopyFileW'),
                    (4301040, 4, 9, 'kernel32.WaitNamedPipeW'),
                    (4301044, 4, 9, 'kernel32.SetConsoleCtrlHandler'),
                    (4301048, 4, 9, 'kernel32.SetConsoleTitleW'),
                    (4301052, 4, 9, 'kernel32.ReadConsoleW'),
                    (4301056, 4, 9, 'kernel32.GetVersion'),
                    (4301060, 4, 9, 'kernel32.SetProcessAffinityMask'),
                    (4301064, 4, 9, 'kernel32.ReadFile'),
                    (4301068, 4, 9, 'kernel32.GetConsoleScreenBufferInfo'),
                    (4301072, 4, 9, 'kernel32.MultiByteToWideChar'),
                    (4301076, 4, 9, 'kernel32.GetComputerNameW'),
                    (4301080, 4, 9, 'kernel32.DeleteFileW'),
                    (4301084, 4, 9, 'kernel32.CreateFileW'),
                    (4301088, 4, 9, 'kernel32.GetSystemDirectoryW'),
                    (4301092, 4, 9, 'kernel32.FindResourceW'),
                    (4301096, 4, 9, 'kernel32.LoadLibraryExW'),
                    (4301100, 4, 9, 'kernel32.FormatMessageA'),
                    (4301104, 4, 9, 'kernel32.GetTickCount'),
                    (4301108, 4, 9, 'kernel32.CloseHandle'),
                    (4301112, 4, 9, 'kernel32.WriteFile'),
                    (4301116, 4, 9, 'kernel32.SizeofResource'),
                    (4301120, 4, 9, 'kernel32.LoadResource'),
                    (4301124, 4, 9, 'kernel32.Sleep'),
                    (4301128, 4, 9, 'kernel32.WaitForSingleObject'),
                    (4301132, 4, 9, 'kernel32.SetEndOfFile'),
                    (4301136, 4, 9, 'kernel32.SetEvent'),
                    (4301140, 4, 9, 'kernel32.SetLastError'),
                    (4301144, 4, 9, 'kernel32.GetLastError'),
                    (4301148, 4, 9, 'kernel32.GetCurrentProcess'),
                    (4301152, 4, 9, 'kernel32.FreeLibrary'),
                    (4301156, 4, 9, 'kernel32.LockResource'),
                    (4301160, 4, 9, 'kernel32.SetPriorityClass'),
                    (4301164, 4, 9, 'kernel32.GetModuleFileNameW'),
                    (4301168, 4, 9, 'kernel32.GetCommandLineW'),
                    (4301172, 4, 9, 'kernel32.GetModuleHandleW'),
                    (4301176, 4, 9, 'kernel32.LoadLibraryW'),
                    (4301180, 4, 9, 'kernel32.GetStdHandle'),
                    (4301184, 4, 9, 'kernel32.GetFileType'),
                    (4301188, 4, 9, 'kernel32.LocalFree'),
                    (4301192, 4, 9, 'kernel32.LocalAlloc'),
                    (4301196, 4, 9, 'kernel32.GetProcAddress'),
                    (4301200, 4, 9, 'kernel32.FreeEnvironmentStringsW'),
                    (4301204, 4, 9, 'kernel32.LCMapStringW'),
                    (4301208, 4, 9, 'kernel32.OutputDebugStringW'),
                    (4301212, 4, 9, 'kernel32.HeapSize'),
                    (4301216, 4, 9, 'kernel32.HeapReAlloc'),
                    (4301220, 4, 9, 'kernel32.SetFilePointerEx'),
                    (4301224, 4, 9, 'kernel32.WriteConsoleW'),
                    (4301228, 4, 9, 'kernel32.GetEnvironmentVariableW'),
                    (4301232, 4, 9, 'kernel32.RaiseException'),
                    (4301236, 4, 9, 'kernel32.LoadLibraryExA'),
                    (4301240, 4, 9, 'kernel32.EncodePointer'),
                    (4301244, 4, 9, 'kernel32.DecodePointer'),
                    (4301248, 4, 9, 'kernel32.ExitProcess'),
                    (4301252, 4, 9, 'kernel32.GetModuleHandleExW'),
                    (4301256, 4, 9, 'kernel32.WideCharToMultiByte'),
                    (4301260, 4, 9, 'kernel32.HeapFree'),
                    (4301264, 4, 9, 'kernel32.HeapAlloc'),
                    (4301268, 4, 9, 'kernel32.GetConsoleMode'),
                    (4301272, 4, 9, 'kernel32.ReadConsoleInputA'),
                    (4301276, 4, 9, 'kernel32.SetConsoleMode'),
                    (4301280, 4, 9, 'kernel32.EnterCriticalSection'),
                    (4301284, 4, 9, 'kernel32.LeaveCriticalSection'),
                    (4301288, 4, 9, 'kernel32.SetStdHandle'),
                    (4301292, 4, 9, 'kernel32.CreateThread'),
                    (4301296, 4, 9, 'kernel32.GetCurrentThreadId'),
                    (4301300, 4, 9, 'kernel32.ExitThread'),
                    (4301304, 4, 9, 'kernel32.IsDebuggerPresent'),
                    (4301308, 4, 9, 'kernel32.IsProcessorFeaturePresent'),
                    (4301312, 4, 9, 'kernel32.GetStringTypeW'),
                    (4301316, 4, 9, 'kernel32.IsValidCodePage'),
                    (4301320, 4, 9, 'kernel32.GetACP'),
                    (4301324, 4, 9, 'kernel32.GetOEMCP'),
                    (4301328, 4, 9, 'kernel32.GetCPInfo'),
                    (4301332, 4, 9, 'kernel32.DeleteCriticalSection'),
                    (4301336, 4, 9, 'kernel32.UnhandledExceptionFilter'),
                    (4301340, 4, 9, 'kernel32.SetUnhandledExceptionFilter'),
                    (4301344, 4, 9, 'kernel32.InitializeCriticalSectionAndSpinCount'),
                    (4301348, 4, 9, 'kernel32.TerminateProcess'),
                    (4301352, 4, 9, 'kernel32.TlsAlloc'),
                    (4301356, 4, 9, 'kernel32.TlsGetValue'),
                    (4301360, 4, 9, 'kernel32.TlsSetValue'),
                    (4301364, 4, 9, 'kernel32.TlsFree'),
                    (4301368, 4, 9, 'kernel32.GetStartupInfoW'),
                    (4301372, 4, 9, 'kernel32.GetProcessHeap'),
                    (4301376, 4, 9, 'kernel32.FlushFileBuffers'),
                    (4301380, 4, 9, 'kernel32.GetConsoleCP'),
                    (4301384, 4, 9, 'kernel32.RtlUnwind'),
                    (4301388, 4, 9, 'kernel32.QueryPerformanceCounter'),
                    (4301392, 4, 9, 'kernel32.GetSystemTimeAsFileTime'),
                    (4301396, 4, 9, 'kernel32.GetEnvironmentStringsW'),
                    (4300976, 4, 9, 'comdlg32.PrintDlgW'),
                    (4300800, 4, 9, 'advapi32.LsaClose'),
                    (4300804, 4, 9, 'advapi32.CreateProcessAsUserW'),
                    (4300808, 4, 9, 'advapi32.CryptHashData'),
                    (4300812, 4, 9, 'advapi32.CryptCreateHash'),
                    (4300816, 4, 9, 'advapi32.CryptDecrypt'),
                    (4300820, 4, 9, 'advapi32.CryptEncrypt'),
                    (4300824, 4, 9, 'advapi32.CryptImportKey'),
                    (4300828, 4, 9, 'advapi32.CryptExportKey'),
                    (4300832, 4, 9, 'advapi32.CryptDestroyKey'),
                    (4300836, 4, 9, 'advapi32.CryptDeriveKey'),
                    (4300840, 4, 9, 'advapi32.CryptGenKey'),
                    (4300844, 4, 9, 'advapi32.CryptReleaseContext'),
                    (4300848, 4, 9, 'advapi32.CryptAcquireContextW'),
                    (4300852, 4, 9, 'advapi32.StartServiceW'),
                    (4300856, 4, 9, 'advapi32.QueryServiceStatus'),
                    (4300860, 4, 9, 'advapi32.OpenServiceW'),
                    (4300864, 4, 9, 'advapi32.OpenSCManagerW'),
                    (4300868, 4, 9, 'advapi32.DeleteService'),
                    (4300872, 4, 9, 'advapi32.CreateServiceW'),
                    (4300876, 4, 9, 'advapi32.ControlService'),
                    (4300880, 4, 9, 'advapi32.CloseServiceHandle'),
                    (4300884, 4, 9, 'advapi32.OpenProcessToken'),
                    (4300888, 4, 9, 'advapi32.LsaEnumerateAccountRights'),
                    (4300892, 4, 9, 'advapi32.LsaOpenPolicy'),
                    (4300896, 4, 9, 'advapi32.LsaFreeMemory'),
                    (4300900, 4, 9, 'advapi32.SetSecurityInfo'),
                    (4300904, 4, 9, 'advapi32.GetSecurityInfo'),
                    (4300908, 4, 9, 'advapi32.LookupPrivilegeValueW'),
                    (4300912, 4, 9, 'advapi32.AddAccessAllowedAce'),
                    (4300916, 4, 9, 'advapi32.GetAce'),
                    (4300920, 4, 9, 'advapi32.AddAce'),
                    (4300924, 4, 9, 'advapi32.InitializeAcl'),
                    (4300928, 4, 9, 'advapi32.GetLengthSid'),
                    (4300932, 4, 9, 'advapi32.FreeSid'),
                    (4300936, 4, 9, 'advapi32.AllocateAndInitializeSid'),
                    (4300940, 4, 9, 'advapi32.SetTokenInformation'),
                    (4300944, 4, 9, 'advapi32.GetTokenInformation'),
                    (4300948, 4, 9, 'advapi32.RegSetValueExW'),
                    (4300952, 4, 9, 'advapi32.RegQueryValueExW'),
                    (4300956, 4, 9, 'advapi32.RegOpenKeyExW'),
                    (4300960, 4, 9, 'advapi32.RegOpenKeyW'),
                    (4300964, 4, 9, 'advapi32.RegCreateKeyW'),
                    (4300968, 4, 9, 'advapi32.RegCloseKey'),
                    (4371208, 4, 9, 'user32.GetSysColorBrush'),
                    (4371212, 4, 9, 'user32.SetCursor'),
                    (4371216, 4, 9, 'user32.LoadCursorW'),
                    (4371220, 4, 9, 'user32.LoadStringW'),
                    (4371224, 4, 9, 'user32.SetWindowTextW'),
                    (4371228, 4, 9, 'user32.GetDlgItem'),
                    (4371232, 4, 9, 'user32.EndDialog'),
                    (4371236, 4, 9, 'user32.DialogBoxIndirectParamW'),
                    (4371240, 4, 9, 'user32.InflateRect'),
                    (4371244, 4, 9, 'user32.SendMessageW'),
                    (4371180, 4, 9, 'gdi32.EndPage'),
                    (4371184, 4, 9, 'gdi32.EndDoc'),
                    (4371188, 4, 9, 'gdi32.StartDocW'),
                    (4371192, 4, 9, 'gdi32.SetMapMode'),
                    (4371196, 4, 9, 'gdi32.GetDeviceCaps'),
                    (4371200, 4, 9, 'gdi32.StartPage')])

        failed = False

        for iloc in vw.getImports():
            if iloc not in imps:
                logger.critical(f'{iloc} was not found in the predefined imports. A new one?')
                failed = True

        oimp = set(vw.getImports())
        for iloc in imps:
            if iloc not in oimp:
                logger.critical(f'The imports are missing {iloc}.')
                failed = True

        if failed:
            self.fail('Please see test logs for import test failures')

    def test_pe_pe32p(self):
        '''
        test that we use the right check for constructing the optional header
        and that the data directories are right
        '''
        file_path = helpers.getTestPath('windows', 'amd64', 'a7712b7c45ae081a1576a387308077f808c666449d1ea9ba680ec410569d476f.file')
        pe = PE.peFromFileName(file_path)
        self.assertTrue(pe.pe32p)
        self.eq(pe.psize, 8)

        self.eq(pe.IMAGE_NT_HEADERS.OptionalHeader.SizeOfStackReserve, 0x100000)
        self.eq(pe.IMAGE_NT_HEADERS.OptionalHeader.SizeOfStackCommit, 0x1000)
        self.eq(pe.IMAGE_NT_HEADERS.OptionalHeader.SizeOfHeapReserve, 0x100000)
        self.eq(pe.IMAGE_NT_HEADERS.OptionalHeader.SizeOfHeapCommit, 0x1000)
        self.eq(pe.IMAGE_NT_HEADERS.OptionalHeader.LoaderFlags, 0)
        self.eq(pe.IMAGE_NT_HEADERS.OptionalHeader.NumberOfRvaAndSizes, 0x10)

        ddir = pe.getDataDirectory(PE.IMAGE_DIRECTORY_ENTRY_COM_DESCRIPTOR)
        self.eq(ddir.VirtualAddress, 0x2008)
        self.eq(ddir.Size, 0x48)

        ddir = pe.getDataDirectory(PE.IMAGE_DIRECTORY_ENTRY_RESOURCE)
        self.eq(ddir.VirtualAddress, 0x9a2000)
        self.eq(ddir.Size, 0x3c78)

    def test_hiaddr_imports(self):
        # Test if imports located at a high relative address are discovered.
        file_path = helpers.getTestPath('windows', 'i386', 'section_has_hi_virtualaddr.exe')
        pe = PE.peFromFileName(file_path)
        import_list = pe.getImports()
        self.assertEqual(len(import_list), 36, "expecting 36 imported functions")
        self.assertEqual(import_list[0][1], "advapi32.dll", "imported function with name 'advapi32.dll' not found")

    def test_mimikatz_segments(self):
        vw = self.vw_mimi
        ans = {
            # name -> (Base, Size, Flags)
            'PE_Header': (0x400000, 0x1000, e_const.MM_READ),
            '.text': (0x401000, 0x71800, e_const.MM_READ | e_const.MM_EXEC),
            '.data': (0x4b6000, 0x4200, e_const.MM_READ | e_const.MM_WRITE),
            '.rdata': (0x473000, 0x42e00, e_const.MM_READ),
            '.reloc': (0x4bf000, 0x6800, e_const.MM_READ),
        }
        for sva, ssize, sname, sfname in vw.getSegments():
            self.assertEqual(ans[sname][0], sva)
            self.assertEqual(ans[sname][1], ssize)
            self.assertEqual(sfname, 'mimikatz')

            mva, msize, flags, mfname = vw.getMemoryMap(sva)
            self.assertEqual(mva, sva)
            self.assertEqual(msize, ssize)
            self.assertEqual(flags, ans[sname][2])
            self.assertEqual(mfname, sfname)

    def test_saved_structures(self):
        ans = [
            0x4001f0,
            0x400218,
            0x400240,
            0x400268,
            0x400290
        ]
        vw = self.vw_psexec
        for va in ans:
            loc = vw.getLocation(va)
            self.assertIsNotNone(loc)
            self.assertEqual(loc[0], va)
            self.assertEqual(loc[1], 40)
            self.assertEqual(loc[2], viv_con.LOC_STRUCT)
            self.assertEqual(loc[3], 'pe.IMAGE_SECTION_HEADER')

    def test_bad_pe_sectname(self):
        vw = helpers.getTestWorkspace('windows', 'i386', 'HelloSection-err.exe')
        segs = vw.getSegments()
        pesections = [
            # about as blunt and direct of a name as we can get
            "[invalid name] b'\\xfftext\\x00\\x00\\x00'",
            '.rdata\x00\x00',
            '.data\x00\x00\x00',
            '.rsrc\x00\x00\x00',
            '.reloc\x00\x00',
        ]
        sections = [x.Name for x in vw.parsedbin.getSections()]
        valids = [x.isNameValid() for x in vw.parsedbin.getSections()]
        self.eq(valids, [False, True, True, True, True])
        self.len(sections, 5)
        for name in pesections:
            self.isin(name, sections)



import vivisect.cli as viv_cli
import vivisect.qt.main as viv_qt_main


def remotemain(appsrv):

    # The "appsrv" is a remote workspace...
    vw = viv_cli.VivCli()
    vw.initWorkspaceClient(appsrv)

    # If we are interactive, lets turn on extended output...
    viv_qt_main.main(vw)



import logging

logger = logging.getLogger(__name__)


class CobraDispatchMethod:
    '''
    implements use of async cobra calls
    '''
    def __init__(self, dispatcher, methname):
        self.dispatcher = dispatcher
        self.methname = methname

    def __call__(self, *args, **kwargs):
        logger.debug("Calling: %s (%s, %s)", self.methname, repr(args)[:20], repr(kwargs)[:20])
        waiters = []

        try:
            for proxy in self.dispatcher.getCobraProxies():
                waiters.append(getattr(proxy, self.methname)(*args, _cobra_async=True, **kwargs))
            return [waiter.wait() for waiter in waiters]
        except Exception:
            for waiter in waiters:
                if waiter.csock:
                    waiter.csock.trashed = True
            raise
        finally:
            for waiter in waiters:
                if waiter.csock and waiter.csock.pool is not None:
                    waiter.csock.pool.put(waiter.csock)


class CobraDispatcher:
    '''
    class implements logic around making async calls
    to multiple cobra proxies. specifically enforces requeuing
    of socket objects when proxy uses sockpool. list of proxies
    can be a mix of proxies with different settings.
    '''
    def __init__(self, proxies):
        self._cobra_proxies = proxies

    def addCobraProxy(self, proxy):
        self._cobra_proxies.append(proxy)

    def delCobraProxy(self, proxy):
        self._cobra_proxies.remove(proxy)

    def getCobraProxies(self):
        return self._cobra_proxies

    def __getattr__(self, name):
        logger.debug("getattr %s", name)

        if name == "__getinitargs__":
            raise AttributeError()

        self._cobra_methods = self._cobra_proxies[0]._cobra_methods
        self._cobra_name = self._cobra_proxies[0]._cobra_name
        # Handle methods
        if self._cobra_methods.get(name, False):
            meth = CobraDispatchMethod(self, name)
            setattr(self, name, meth)
            return meth




import vstruct
from vstruct.primitives import *

GIF_F_HAS_CMAP  = 0x80
GIF_F_BPP_MASK  = 0x07

GIF_IMG_SEP     = ','

class GIF_FILE_HEADER(vstruct.VStruct):

    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.magic      = v_bytes(size=6)
        self.width      = v_uint16()
        self.height     = v_uint16()
        self.flags      = v_uint8()
        self.bgcolor    = v_uint8()
        self.zero       = v_uint8()

class RGB(vstruct.VStruct):

    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.red    = v_uint8()
        self.green  = v_uint8()
        self.blue   = v_uint8()

class GIF_IMAGE_DESCRIPTOR(vstruct.VStruct):

    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.sep        = v_uint8()
        self.img_left   = v_uint16()
        self.img_top    = v_uint16()
        self.img_width  = v_uint16()
        self.img_height = v_uint16()
        self.flags      = v_uing8()

class GIF8XA(vstruct.VStruct):

    def __init__(self):
        vstruct.VStruct.__init__(self)

    def isValidGif(self):
        if self.header.magic not in ('GIF87a', 'GIF89a'):
            return False
        if self.header.zero != 0:
            return False

    def vsParse(self, bytes, offset):


        # FIXME this is not functional yet...

        self.vsClearFields()
        self.header = GIF_FILE_HEADER()
        offset = self.header.vsParse(bytes, offset)

        # Do we have a global color table?
        if self.header.flags & GIF_F_HAS_CMAP:
            bits_per_pixel = (self.header.flags & GIF_F_BPP_MASK) + 1
            self.gct = vstruct.VStruct()
            for i in range(2**bits_per_pixel):
                self.gct.vsAddField('color%d' % i, RGB())

            offset = self.gct.vsParse(bytes, offset)

        self.images = vstruct.VStruct()

        imgidx = 0
        while bytes[offset] == GIF_IMG_SEP:
            img = vstruct.VStruct()

            img.descriptor = GIF_IMAGE_DESCRIPTOR()
            offset = img.descriptor.vsParse(bytes, offset)

            if img.descriptor.flags & GIF_F_HAS_CMAP:
                bits_per_pixel = (img.descriptor.flags & GIF_F_BPP_MASK) + 1
                img.cmap = vstruct.VStruct()

                for i in range(2**bits_per_pixel):
                    img.cmap.vsAddField('color%d' % i, RGB())

                offset = img.cmap.vsParse(bytes, offset)




import os
import json
import logging

import envi.exc as e_exc
import envi.config as e_config

logger = logging.getLogger(__name__)

def symCacheHashFromPe(pe):
    checksum = pe.IMAGE_NT_HEADERS.OptionalHeader.CheckSum
    codesize = pe.IMAGE_NT_HEADERS.OptionalHeader.SizeOfCode
    timestamp = pe.IMAGE_NT_HEADERS.FileHeader.TimeDateStamp
    return 'pe.%.8x.%.8x.%.8x' % (timestamp,checksum,codesize)

def symCacheHashFromElf(elf):
    pass #FIXME

class SymbolCache:
    '''
    A SymbolCache is a location where pre-parsed symbols for
    a given file's "symcache hash" are stored.  This allows faster
    loading of symbol information on re-load etc...

    Also, this object's API *must* be compatible with cobra
    shared object API to allow "symbol servers" to be cobra
    shared instances of SymbolCache objects.
    '''
    def __init__(self, dirname=None):

        if dirname is None:
            dirname = e_config.gethomedir('.envi', 'symcache')

        if not os.path.isdir(dirname):
            os.makedirs(dirname)

        self._sym_cachedir = os.path.abspath(dirname)

    def setCacheSyms(self, vhash, symcache):
        '''
        Save a set of symbol cache tuples to the symbol cache.

        Example:
            tups = [ ( rva, size, 'wootfunc', SYMSTOR_SYM_FUNCTION), ]
            cache = SymbolCache()
            cache.setCacheSyms( vhash, tups )
        '''
        cachefile = os.path.join(self._sym_cachedir, vhash)

        abspath = os.path.abspath(cachefile)
        if not abspath.startswith(self._sym_cachedir):
            raise e_exc.InvalidSymbolCache(vhash)

        # FIXME check input path
        with open(cachefile, 'w', encoding='utf-8') as fd:
            json.dump(symcache, fd)

    def getCacheSyms(self, vhash):
        '''
        Retrieve a list of symbol tuples ( rva, size, name, symtype )
        or None if the symbol cache doesn't have the given file hash.

        Example:
            cache = SymbolCache()
            for rva, size, name, stype in cache.getCacheSyms():
                dostuff()
        '''
        cachefile = os.path.join(self._sym_cachedir, vhash)

        abspath = os.path.abspath(cachefile)
        if not abspath.startswith(self._sym_cachedir):
            raise e_exc.InvalidSymbolCache(vhash)

        if not os.path.isfile(cachefile):
            return None

        try:
            with open(cachefile, 'r', encoding='utf-8') as fd:
                return json.load(fd)
        except Exception as e:
            logger.warning('Failed to load cachefile: %s', e)
            return None


class SymbolCachePath:

    def __init__(self, path):
        self.symcaches = []

        for path in path.split(';'):

            if os.path.isdir(path):
                self.symcaches.append(SymbolCache(dirname=path))
                continue

            if path.startswith('cobra://') or path.startswith('cobrassl://'):
                import cobra
                self.symcaches.append(cobra.CobraProxy(path))
                continue

            #if path.startswith('http://') or path.startswith('https://'):
                #FIXME

    def getCacheSyms(self, symhash):

        for symcache in self.symcaches:
            ret = symcache.getCacheSyms(symhash)
            if ret is not None:
                return ret

    def setCacheSyms(self, symhash, symcache):
        if self.symcaches:
            self.symcaches[0].setCacheSyms(symhash, symcache)



import vtrace
import vtrace.util

import vtrace.tests


class EnvitoolsTests(vtrace.tests.VtraceProcessTest):
    def test_emulatorFromTrace(self):
        emu = vtrace.util.emuFromTrace(self.trace)

        ctx = self.trace.getRegisterContext()
        for _, idx in ctx.getRegisterNameIndexes():
            self.assertEqual(self.trace.getRegister(idx), emu.getRegister(idx))

        emu.stepi()

    def test_workspaceFromTrace(self):
        vw = vtrace.util.vwFromTrace(self.trace, collapse=False)

        for mva, _, mperms, mfname in self.trace.getMemoryMaps():
            if mfname in self.trace.getMeta('BadMaps') or mperms == 0:
                continue

            # test that the file mappings exist for each (which means segments have to be correct)
            filename = vw.getFileByVa(mva)
            self.assertTrue(filename in mfname)
